"""
The Job Shop Scheduling Problem is a good example for optimization tasks.

The JSSP is one of the most well-known combinatorial optimization tasks.
Here we provide the well-known and often-used set of benchmark instances
(:mod:`~moptipy.examples.jssp.instance`) for this problem. Moreover, we
also run a fully-fledged example experiment applying many metaheuristics
to eight selected JSSP instances in module
:mod:`~moptipy.examples.jssp.experiment` and evaluate the data gathered from
this experiment in :mod:`~moptipy.examples.jssp.evaluation`. Be aware that
actually executing this experiment will take a very long time. However, you
can also find the complete experiment, every single algorithm used, and all
conclusions drawn with in-depth discussions in our online book "Optimization
Algorithms" at https://thomasweise.github.io/oa.

Anyway, here you have many JSSP benchmark instances, a useful search space,
and the common objective function. You can use them for your own research, if
you want: The JSSP benchmark instances can be loaded from a package-internal
resource by name using module :mod:`~moptipy.examples.jssp.instance`.
Solutions to a JSSP instances are :class:`~moptipy.examples.jssp.gantt.Gantt`
charts, i.e., diagrams which assign each operation of a job to a machine and
time slot. The :class:`~moptipy.examples.jssp.gantt_space.GanttSpace` is a
space operating on such charts and the
:class:`~moptipy.examples.jssp.makespan.Makespan` is the typical objective
function when solving JSSPs. Gantt charts can be encoded as permutations
with repetitions using the :mod:`~moptipy.examples.jssp.ob_encoding`.

1. Ronald Lewis Graham, Eugene Leighton Lawler, Jan Karel Lenstra, Alexander
   Hendrik George Rinnooy Kan. Optimization and Approximation in Deterministic
   Sequencing and Scheduling: A Survey. *Annals of Discrete Mathematics*
   5:287-326, 1979. doi: https://doi.org/10.1016/S0167-5060(08)70356-X.
   https://ir.cwi.nl/pub/18052/18052A.pdf.
2. Eugene Leighton Lawler, Jan Karel Lenstra, Alexander Hendrik George Rinnooy
   Kan, and David B. Shmoys. Sequencing and Scheduling: Algorithms and
   Complexity. Chapter 9 in Stephen C. Graves, Alexander Hendrik George
   Rinnooy Kan, and Paul H. Zipkin, editors, *Handbook of Operations Research
   and Management Science,* volume IV: *Production Planning and Inventory,*
   1993, pages 445-522. Amsterdam, The Netherlands: North-Holland Scientific
   Publishers Ltd. doi: https://doi.org/10.1016/S0927-0507(05)80189-6.
   http://alexandria.tue.nl/repository/books/339776.pdf.
3. Eugene Leighton Lawler. Recent Results in the Theory of Machine Scheduling.
   Chapter 8 in Achim Bachem, Bernhard Korte, and Martin Grötschel, editors,
   *Math Programming: The State of the Art,* 1982, pages 202-234.
   Bonn, Germany/New York, NY, USA: Springer-Verlag GmbH.
   ISBN: 978-3-642-68876-8. doi: https://doi.org/10.1007/978-3-642-68874-4_9.
4. Éric D. Taillard. Benchmarks for Basic Scheduling Problems. *European
   Journal of Operational Research (EJOR)* 64(2):278-285, January 1993.
   doi: https://doi.org/10.1016/0377-2217(93)90182-M.
   http://mistic.heig-vd.ch/taillard/articles.dir/Taillard1993EJOR.pdf.
5. Jacek Błażewicz, Wolfgang Domschke, and Erwin Pesch. The Job Shop
   Scheduling Problem: Conventional and New Solution Techniques. *European
   Journal of Operational Research (EJOR)* 93(1):1-33, August 1996.
   doi: https://doi.org/10.1016/0377-2217(95)00362-2.
6. Thomas Weise. *Optimization Algorithms.* 2021-2023. Hefei, Anhui, China:
   Institute of Applied Optimization, School of Artificial Intelligence and
   Big Data, Hefei University. https://thomasweise.github.io/oa
"""


def __lang_setup() -> None:
    """Perform the language setup."""
    from moptipy.utils.lang import DE, EN, ZH  # pylint: disable=C0415

    # the English language strings
    EN.extend({
        "gantt_info": "{gantt.instance.name} ({gantt.instance.jobs} "
                      "jobs \u00D7 {gantt.instance.machines} machines), "
                      "makespan {gantt[:,:,2].max()}",
        "gantt_info_no_ms": "{gantt.instance.name} ({gantt.instance.jobs} "
                            "jobs \u00D7 {gantt.instance.machines} machines)",
        "gantt_info_short": "{gantt.instance.name} / {gantt[:,:,2].max()}",
        "machine": "machine",
        "makespan": "makespan",
    })
    # the German language strings
    DE.extend({
        "gantt_info": "{gantt.instance.name} ({gantt.instance.jobs} "
                      "Jobs \u00D7 {gantt.instance.machines} Maschinen), "
                      "Makespan {gantt[:,:,2].max()}",
        "gantt_info_no_ms": "{gantt.instance.name} ({gantt.instance.jobs} "
                            "Jobs \u00D7 {gantt.instance.machines} Maschinen)",
        "gantt_info_short": "{gantt.instance.name} / {gantt[:,:,2].max()}",
        "machine": "Maschine",
        "makespan": "Makespan",
    })
    # the Chinese language strings
    ZH.extend({
        "gantt_info": "{gantt.instance.name} ({gantt.instance.jobs}份作业"
                      "\u00D7{gantt.instance.machines}台机器), "
                      "最大完工时间{gantt[:,:,2].max()}",
        "gantt_info_no_ms":
            "{gantt.instance.name} ({gantt.instance.jobs}份作业"
            "\u00D7{gantt.instance.machines}台机器) ",
        "gantt_info_short": "{gantt.instance.name} / {gantt[:,:,2].max()}",
        "machine": "机器",
        "makespan": "最大完工时间",
    })


__lang_setup()  # invoke the language setup
del __lang_setup  # delete the language setup routine



"""A process with logging and different search and solution space."""
from typing import Final

from pycommons.io.path import Path
from pycommons.types import type_error

from moptipy.api._process_base import _TIME_IN_NS, _check_log_time
from moptipy.api._process_no_ss import _write_log
from moptipy.api._process_ss import _ProcessSS
from moptipy.api.algorithm import Algorithm
from moptipy.api.encoding import Encoding
from moptipy.api.objective import Objective
from moptipy.api.space import Space
from moptipy.utils.logger import Logger


class _ProcessSSLog(_ProcessSS):
    """A process with logging and different search and solution space."""

    def __init__(self,
                 solution_space: Space,
                 objective: Objective,
                 algorithm: Algorithm,
                 log_file: Path,
                 search_space: Space | None = None,
                 encoding: Encoding | None = None,
                 rand_seed: int | None = None,
                 max_fes: int | None = None,
                 max_time_millis: int | None = None,
                 goal_f: int | float | None = None,
                 log_all_fes: bool = False) -> None:
        """
        Perform the internal initialization. Do not call directly.

        :param solution_space: the solution space.
        :param objective: the objective function
        :param algorithm: the optimization algorithm
        :param search_space: the search space.
        :param encoding: the encoding
        :param log_file: the optional log file
        :param rand_seed: the optional random seed
        :param max_fes: the maximum permitted function evaluations
        :param max_time_millis: the maximum runtime in milliseconds
        :param goal_f: the goal objective value. if it is reached, the
            process is terminated
        :param log_all_fes: should every single FE be logged?
        """
        super().__init__(solution_space=solution_space,
                         objective=objective,
                         algorithm=algorithm,
                         log_file=log_file,
                         search_space=search_space,
                         encoding=encoding,
                         rand_seed=rand_seed,
                         max_fes=max_fes,
                         max_time_millis=max_time_millis,
                         goal_f=goal_f)
        if not isinstance(log_file, str):
            raise type_error(log_file, "log_file", str)
        if not isinstance(log_all_fes, bool):
            raise type_error(log_all_fes, "log_all_fes", bool)
        #: `True` if all FEs are logged, `False` to only log improvements.
        self.__log_all: Final[bool] = log_all_fes
        #: The in-memory log
        self.__log: list[list[int | float]] = []
        #: the quick access to the log appending method
        self.__log_append = self.__log.append

    def evaluate(self, x) -> float | int:
        if self._terminated:
            if self._knows_that_terminated:
                raise ValueError("The process has been terminated and the "
                                 "algorithm knows it.")
            return self._current_best_f

        current_y: Final = self._current_y
        self._g(x, current_y)
        result: Final[int | float] = self._f(current_y)
        self._current_fes = current_fes = self._current_fes + 1
        do_term: bool = current_fes >= self._end_fes
        do_log: bool = self.__log_all
        ctn: int = 0

        if result < self._current_best_f:
            self._last_improvement_fe = current_fes
            self._current_best_f = result
            self.copy(self._current_best_x, x)
            self._current_y = self._current_best_y
            self._current_best_y = current_y
            self._current_time_nanos = ctn = _TIME_IN_NS()
            self._last_improvement_time_nanos = ctn
            do_term = do_term or (result <= self._end_f)
            do_log = True

        if do_log:
            if ctn <= 0:
                self._current_time_nanos = ctn = _TIME_IN_NS()
            self.__log_append([current_fes, ctn, result])

        if do_term:
            self.terminate()

        return result

    def register(self, x, f: int | float) -> None:
        if self._terminated:
            if self._knows_that_terminated:
                raise ValueError("The process has been terminated and the "
                                 "algorithm knows it.")
            return

        self._current_fes = current_fes = self._current_fes + 1
        do_term: bool = current_fes >= self._end_fes
        do_log: bool = self.__log_all
        ctn: int = 0

        if f < self._current_best_f:
            self._last_improvement_fe = current_fes
            self._current_best_f = f
            self.copy(self._current_best_x, x)
            current_y: Final = self._current_y
            self._g(x, current_y)
            self._current_y = self._current_best_y
            self._current_best_y = current_y
            self._current_time_nanos = ctn = _TIME_IN_NS()
            self._last_improvement_time_nanos = ctn
            do_term = do_term or (f <= self._end_f)
            do_log = True

        if do_log:
            if ctn <= 0:
                self._current_time_nanos = ctn = _TIME_IN_NS()
            self.__log_append([current_fes, ctn, f])

        if do_term:
            self.terminate()

    def _check_timing(self) -> None:
        super()._check_timing()
        _check_log_time(self._start_time_nanos, self._current_time_nanos,
                        self.__log)

    def _write_log(self, logger: Logger) -> None:
        _write_log(self.__log, self._start_time_nanos, logger)
        del self.__log
        super()._write_log(logger)

    def __str__(self) -> str:
        return "LoggingProcessWithSearchSpace"



"""
The simulated annealing algorithm with configurable temperature schedule.

A basic randomized local search (:mod:`~moptipy.algorithms.so.rls`) maintains
one interesting solution and derives one new solution from it using the
unary search operator (:class:`~moptipy.api.operators.Op1`). The new solution
replaces the current solution if it is not worse, i.e., better or equally
good. Simulated Annealing is similar to the :mod:`~moptipy.algorithms.so.rls`,
but it sometimes also accepts solutions that are worse than the current one.
It does so with a probability that becomes smaller the worse the new solution
is and also becomes smaller the smaller the current "temperature" is.

Simulated Annealing applies a so-called temperature schedule (see
:mod:`~moptipy.algorithms.modules.temperature_schedule`), which basically is
function that relates the index of the algorithm iteration (i.e., the index
of the current objective function evaluation) to a temperature. It therefore
is a function accepting an integer value as input and returning a float
temperature. This function is usually monotonously decreasing over time,
meaning that the initial "temperature" is high and then becomes smaller. The
algorithm therefore is more likely to accept worse solutions at its beginning,
whereas it behaves basically like a :mod:`~moptipy.algorithms.so.rls` at the
end of its computational budget (if configured correctly).

Simulated Annealing was independently developed by several researchers [1-4].
The idea is inspired by Metropolis' approximation of how annealing can be
simulated [5].

1. Scott Kirkpatrick, C. Daniel Gelatt, Jr., and Mario P. Vecchi. Optimization
   by Simulated Annealing. *Science Magazine.* 220(4598):671-680.
   May 13, 1983. doi: https://doi.org/10.1126/science.220.4598.671.
   https://www.researchgate.net/publication/6026283
2. Vladimír Černý. Thermodynamical Approach to the Traveling Salesman Problem:
   An Efficient Simulation Algorithm. *Journal of Optimization Theory and
   Applications.* 45(1):41-51. January 1985.
   doi: https://doi.org/10.1007/BF00940812.
   http://mkweb.bcgsc.ca/papers/cerny-travelingsalesman.pdf.
3. Dean Jacobs, Jan Prins, Peter Siegel, and Kenneth Wilson. Monte Carlo
   Techniques in Code Optimization. *ACM SIGMICRO Newsletter.* 13(4):143-148.
   December 1982. Also in Proceedings of the 15th Annual Workshop on
   Microprogramming (MICRO 15), October 5-7, 1982, Palo Alto, CA, USA,
   New York, NY, USA: ACM. doi: http://doi.org/10.1145/1014194.800944.
4. Martin Pincus. Letter to the Editor - A Monte Carlo Method for the
   Approximate Solution of Certain Types of Constrained Optimization Problems.
   *Operations Research.* 18(6):1225-1228. November/December 1970.
   doi: https://doi.org/10.1287/opre.18.6.1225.
5. Nicholas Metropolis, Arianna W. Rosenbluth, Marshall Nicholas Rosenbluth,
   Augusta H. Teller, Edward Teller. Equation of State Calculations by Fast
   Computing Machines. *The Journal of Chemical Physics*. 21(6):1087-1092.
   June 1953. doi: https://doi.org/10.1063/1.1699114.
   http://scienze-como.uninsubria.it/bressanini/montecarlo-history/mrt2.pdf.
"""
from math import exp
from typing import Callable, Final

from numpy.random import Generator
from pycommons.types import type_error

from moptipy.algorithms.modules.temperature_schedule import TemperatureSchedule
from moptipy.api.algorithm import Algorithm1
from moptipy.api.operators import Op0, Op1
from moptipy.api.process import Process
from moptipy.utils.logger import KeyValueLogSection


# start book
class SimulatedAnnealing(Algorithm1):
    """Simulated Annealing is an RLS sometimes accepting worsening moves."""

    def solve(self, process: Process) -> None:
        """
        Apply the simulated annealing algorithm to an optimization problem.

        :param process: the black-box process object
        """
        # Create records for old and new point in the search space.
        best_x = process.create()  # record for best-so-far solution
        new_x = process.create()  # record for new solution
        # Obtain the random number generator.
        random: Final[Generator] = process.get_random()

        # Put function references in variables to for faster calls.
        temperature: Final[Callable[[int], float]] = self.schedule.temperature
        r01: Final[Callable[[], float]] = random.random  # random from [0, 1)
        evaluate: Final[Callable] = process.evaluate  # the objective
        op1: Final[Callable] = self.op1.op1  # the unary operator
        should_terminate: Final[Callable] = process.should_terminate

        # Start at a random point in the search space and evaluate it.
        self.op0.op0(random, best_x)  # Create 1 solution randomly and
        best_f: int | float = evaluate(best_x)  # evaluate it.
        tau: int = 0  # The iteration index, needs to be 0 at first cmp.

        while not should_terminate():  # Until we need to quit...
            op1(random, new_x, best_x)  # new_x = neighbor of best_x
            new_f: int | float = evaluate(new_x)
            if (new_f <= best_f) or (  # Accept if <= or if SA criterion
                    r01() < exp((best_f - new_f) / temperature(tau))):
                best_f = new_f  # Store its objective value.
                best_x, new_x = new_x, best_x  # Swap best and new.
            tau = tau + 1  # Step the iteration index.
# end book

    def __init__(self, op0: Op0, op1: Op1,
                 schedule: TemperatureSchedule) -> None:
        """
        Create the simulated annealing algorithm.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param schedule: the temperature schedule to use
        """
        super().__init__(f"sa_{schedule}", op0, op1)
        if not isinstance(schedule, TemperatureSchedule):
            raise type_error(schedule, "schedule", TemperatureSchedule)
        #: the temperature schedule
        self.schedule: Final[TemperatureSchedule] = schedule

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of the simulated annealing algorithm.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope("ts") as ts:
            self.schedule.log_parameters_to(ts)

    def initialize(self) -> None:
        """Initialize the algorithm."""
        super().initialize()
        self.schedule.initialize()



"""Violin plots for end results."""
from typing import Any, Callable, Final, Iterable, cast

import matplotlib.collections as mc  # type: ignore
from matplotlib.axes import Axes  # type: ignore
from matplotlib.figure import Figure  # type: ignore
from matplotlib.lines import Line2D  # type: ignore
from pycommons.io.console import logger
from pycommons.types import type_error

import moptipy.utils.plot_defaults as pd
import moptipy.utils.plot_utils as pu
from moptipy.evaluation.axis_ranger import AxisRanger
from moptipy.evaluation.base import F_NAME_SCALED
from moptipy.evaluation.end_results import EndResult
from moptipy.evaluation.end_results import getter as end_result_getter
from moptipy.utils.lang import Lang


def plot_end_results(
        end_results: Iterable[EndResult],
        figure: Axes | Figure,
        dimension: str = F_NAME_SCALED,
        y_axis: AxisRanger | Callable[[str], AxisRanger] =
        AxisRanger.for_axis,
        distinct_colors_func: Callable[[int], Any] = pd.distinct_colors,
        importance_to_line_width_func: Callable[[int], float] =
        pd.importance_to_line_width,
        importance_to_font_size_func: Callable[[int], float] =
        pd.importance_to_font_size,
        y_grid: bool = True,
        x_grid: bool = True,
        x_label: None | str | Callable[[str], str] = Lang.translate,
        x_label_inside: bool = True,
        x_label_location: float = 1.0,
        y_label: None | str | Callable[[str], str] = Lang.translate,
        y_label_inside: bool = True,
        y_label_location: float = 0.5,
        legend_pos: str = "best",
        instance_sort_key: Callable[[str], Any] = lambda x: x,
        algorithm_sort_key: Callable[[str], Any] = lambda x: x,
        instance_namer: Callable[[str], str] = lambda x: x,
        algorithm_namer: Callable[[str], str] = lambda x: x) -> Axes:
    """
    Plot a set of end result boxes/violins functions into one chart.

    In this plot, we combine two visualizations of data distributions: box
    plots in the foreground and violin plots in the background.

    The box plots show you the median, the 25% and 75% quantiles, the 95%
    confidence interval around the median (as notches), the 5% and 95%
    quantiles (as whiskers), the arithmetic mean (as triangle), and the
    outliers on both ends of the spectrum. This allows you also to compare
    data from different distributions rather comfortably, as you can, e.g.,
    see whether the confidence intervals overlap.

    The violin plots in the background are something like smoothed-out,
    vertical, and mirror-symmetric histograms. They give you a better
    impression about shape and modality of the distribution of the results.

    :param end_results: the iterable of end results
    :param figure: the figure to plot in
    :param dimension: the dimension to display
    :param y_axis: the y_axis ranger
    :param distinct_colors_func: the function returning the palette
    :param importance_to_line_width_func: the function converting importance
        values to line widths
    :param importance_to_font_size_func: the function converting importance
        values to font sizes
    :param y_grid: should we have a grid along the y-axis?
    :param x_grid: should we have a grid along the x-axis?
    :param x_label: a callable returning the label for the x-axis, a label
        string, or `None` if no label should be put
    :param x_label_inside: put the x-axis label inside the plot (so that
        it does not consume additional vertical space)
    :param x_label_location: the location of the x-label
    :param y_label: a callable returning the label for the y-axis, a label
        string, or `None` if no label should be put
    :param y_label_inside: put the y-axis label inside the plot (so that it
        does not consume additional horizontal space)
    :param y_label_location: the location of the y-label
    :param legend_pos: the legend position
    :param instance_sort_key: the sort key function for instances
    :param algorithm_sort_key: the sort key function for algorithms
    :param instance_namer: the name function for instances receives an
        instance ID and returns an instance name; default=identity function
    :param algorithm_namer: the name function for algorithms receives an
        algorithm ID and returns an instance name; default=identity function
    :returns: the axes object to allow you to add further plot elements
    """
    # Before doing anything, let's do some type checking on the parameters.
    # I want to ensure that this function is called correctly before we begin
    # to actually process the data. It is better to fail early than to deliver
    # some incorrect results.
    if not isinstance(end_results, Iterable):
        raise type_error(end_results, "end_results", Iterable)
    if not isinstance(figure, Axes | Figure):
        raise type_error(figure, "figure", (Axes, Figure))
    if not isinstance(dimension, str):
        raise type_error(dimension, "dimension", str)
    if not callable(distinct_colors_func):
        raise type_error(
            distinct_colors_func, "distinct_colors_func", call=True)
    if not callable(importance_to_line_width_func):
        raise type_error(importance_to_line_width_func,
                         "importance_to_line_width_func", call=True)
    if not callable(importance_to_font_size_func):
        raise type_error(importance_to_font_size_func,
                         "importance_to_font_size_func", call=True)
    if not isinstance(y_grid, bool):
        raise type_error(y_grid, "y_grid", bool)
    if not isinstance(x_grid, bool):
        raise type_error(x_grid, "x_grid", bool)
    if not ((x_label is None) or callable(x_label)
            or isinstance(x_label, str)):
        raise type_error(x_label, "x_label", (str, None), call=True)
    if not isinstance(x_label_inside, bool):
        raise type_error(x_label_inside, "x_label_inside", bool)
    if not isinstance(x_label_location, float):
        raise type_error(x_label_location, "x_label_location", float)
    if not ((y_label is None) or callable(y_label)
            or isinstance(y_label, str)):
        raise type_error(y_label, "y_label", (str, None), call=True)
    if not isinstance(y_label_inside, bool):
        raise type_error(y_label_inside, "y_label_inside", bool)
    if not isinstance(y_label_location, float):
        raise type_error(y_label_location, "y_label_location", float)
    if not isinstance(legend_pos, str):
        raise type_error(legend_pos, "legend_pos", str)
    if not callable(instance_sort_key):
        raise type_error(instance_sort_key, "instance_sort_key", call=True)
    if not callable(algorithm_sort_key):
        raise type_error(algorithm_sort_key, "algorithm_sort_key", call=True)
    if not callable(instance_namer):
        raise type_error(instance_namer, "instance_namer", call=True)
    if not callable(algorithm_namer):
        raise type_error(algorithm_namer, "algorithm_namer", call=True)

    getter: Final[Callable[[EndResult], int | float]] \
        = end_result_getter(dimension)
    logger(f"now plotting end violins for dimension {dimension}.")

    if callable(y_axis):
        y_axis = y_axis(dimension)
    if not isinstance(y_axis, AxisRanger):
        raise type_error(y_axis, f"y_axis for {dimension}", AxisRanger)

    # instance -> algorithm -> values
    data: dict[str, dict[str, list[int | float]]] = {}
    algo_set: set[str] = set()

    # We now collect instances, the algorithms, and the measured values.
    for res in end_results:
        if not isinstance(res, EndResult):
            raise type_error(res, "violin plot element", EndResult)

        algo_set.add(res.algorithm)

        per_inst_data: dict[str, list[int | float]]
        if res.instance not in data:
            data[res.instance] = per_inst_data = {}
        else:
            per_inst_data = data[res.instance]
        inst_algo_data: list[int | float]
        if res.algorithm not in per_inst_data:
            per_inst_data[res.algorithm] = inst_algo_data = []
        else:
            inst_algo_data = per_inst_data[res.algorithm]

        value: int | float = getter(res)
        if not isinstance(value, int | float):
            raise type_error(value, "value", (int, float))
        inst_algo_data.append(value)
        y_axis.register_value(value)

    # We now know the number of instances and algorithms and have the data in
    # the hierarchical structure instance->algorithms->values.
    n_instances: Final[int] = len(data)
    n_algorithms: Final[int] = len(algo_set)
    if (n_instances <= 0) or (n_algorithms <= 0):
        raise ValueError("Data cannot be empty but found "
                         f"{n_instances} and {n_algorithms}.")
    algorithms: Final[tuple[str, ...]] = \
        tuple(sorted(algo_set, key=algorithm_sort_key))
    logger(f"- {n_algorithms} algorithms ({algorithms}) "
           f"and {n_instances} instances ({data.keys()}).")

    # compile the data
    inst_algos: list[tuple[str, list[str]]] = []
    plot_data: list[list[int | float]] = []
    plot_algos: list[str] = []
    instances: Final[list[str]] = sorted(data.keys(), key=instance_sort_key)
    for inst in instances:
        per_inst_data = data[inst]
        algo_names: list[str] = sorted(per_inst_data.keys(),
                                       key=algorithm_sort_key)
        plot_algos.extend(algo_names)
        inst_algos.append((inst, algo_names))
        for algo in algo_names:
            inst_algo_data = per_inst_data[algo]
            inst_algo_data.sort()
            plot_data.append(inst_algo_data)

    # compute the violin positions
    n_bars: Final[int] = len(plot_data)
    if n_bars < max(n_instances, n_algorithms):
        raise ValueError(f"Huh? {n_bars}, {n_instances}, {n_algorithms}")
    bar_positions: Final[tuple[int, ...]] = \
        tuple(range(1, len(plot_data) + 1))

    # Now we got all instances and all algorithms and know the axis ranges.
    font_size_0: Final[float] = importance_to_font_size_func(0)

    # set up the graphics area
    axes: Final[Axes] = pu.get_axes(figure)
    axes.tick_params(axis="y", labelsize=font_size_0)
    axes.tick_params(axis="x", labelsize=font_size_0)

    z_order: int = 0

    # draw the grid
    grid_lwd: int | float | None = None
    if y_grid:
        grid_lwd = importance_to_line_width_func(-1)
        z_order += 1
        axes.grid(axis="y", color=pd.GRID_COLOR, linewidth=grid_lwd,
                  zorder=z_order)

    x_axis: Final[AxisRanger] = AxisRanger(
        chosen_min=0.5, chosen_max=bar_positions[-1] + 0.5)

    # manually add x grid lines between instances
    if x_grid and (n_instances > 1) and (n_algorithms > 1):
        if not grid_lwd:
            grid_lwd = importance_to_line_width_func(-1)
        counter: int = 0
        for key in inst_algos:
            if counter > 0:
                z_order += 1
                axes.axvline(x=counter + 0.5,
                             color=pd.GRID_COLOR,
                             linewidth=grid_lwd,
                             zorder=z_order)
            counter += len(key[1])

    y_axis.apply(axes, "y")
    x_axis.apply(axes, "x")

    violin_width: Final[float] = 3 / 4
    z_order += 1
    violins: Final[dict[str, Any]] = axes.violinplot(
        dataset=plot_data, positions=bar_positions, vert=True,
        widths=violin_width, showmeans=False, showextrema=False,
        showmedians=False)

    # fix the algorithm colors
    unique_colors: Final[tuple[Any]] = distinct_colors_func(n_algorithms)
    algo_colors: Final[dict[str, tuple[float, float, float]]] = {}
    for i, algo in enumerate(algorithms):
        algo_colors[algo] = unique_colors[i]

    bodies: Final[list[mc.PolyCollection]] = violins["bodies"]
    use_colors: Final[list[tuple[float, float, float]]] = []
    counter = 0
    for key in inst_algos:
        for algo in key[1]:
            z_order += 1
            bd = bodies[counter]
            color = algo_colors[algo]
            use_colors.append(color)
            bd.set_edgecolor("none")
            bd.set_facecolor(color)
            bd.set_alpha(0.6666666666)
            counter += 1
            bd.set_zorder(z_order)

    z_order += 1
    boxes_bg: Final[dict[str, Any]] = axes.boxplot(
        x=plot_data, positions=bar_positions, widths=violin_width,
        showmeans=True, patch_artist=False, notch=True, vert=True,
        whis=(5.0, 95.0), manage_ticks=False, zorder=z_order)
    z_order += 1
    boxes_fg: Final[dict[str, Any]] = axes.boxplot(
        x=plot_data, positions=bar_positions, widths=violin_width,
        showmeans=True, patch_artist=False, notch=True, vert=True,
        whis=(5.0, 95.0), manage_ticks=False, zorder=z_order)

    for tkey in ("cmeans", "cmins", "cmaxes", "cbars", "cmedians",
                 "cquantiles"):
        if tkey in violins:
            violins[tkey].set_color("none")

    lwd_fg = importance_to_line_width_func(0)
    lwd_bg = importance_to_line_width_func(1)

    for bid, boxes in enumerate([boxes_bg, boxes_fg]):
        for tkey in ("boxes", "medians", "whiskers", "caps", "fliers",
                     "means"):
            if tkey not in boxes:
                continue
            polys: list[Line2D] = boxes[tkey]
            for line in polys:
                xdata = cast(list, line.get_xdata(True))
                if len(xdata) <= 0:
                    line.remove()
                    continue
                index = int(max(xdata)) - 1
                thecolor: str | tuple[float, float, float] = \
                    "white" if bid == 0 else use_colors[index]
                width = lwd_bg if bid == 0 else lwd_fg
                line.set_solid_joinstyle("round")
                line.set_solid_capstyle("round")
                line.set_color(thecolor)
                line.set_linewidth(width)
                line.set_markeredgecolor(thecolor)
                line.set_markerfacecolor("none")
                line.set_markeredgewidth(width)
                z_order = z_order + 1
                line.set_zorder(z_order)

    # compute the labels for the x-axis
    labels_str: list[str] = []
    labels_x: list[float] = []
    needs_legend: bool = False

    counter = 0
    if n_instances > 1:
        # use only the instances as labels
        for key in inst_algos:
            current = counter
            counter += len(key[1])
            labels_str.append(instance_namer(key[0]))
            labels_x.append(0.5 * (bar_positions[current]
                                   + bar_positions[counter - 1]))
        needs_legend = (n_algorithms > 1)
    elif n_algorithms > 1:
        # only use algorithms as key
        for key in inst_algos:
            for algo in key[1]:
                labels_str.append(algorithm_namer(algo))
                labels_x.append(bar_positions[counter])
                counter += 1

    if labels_str:
        axes.set_xticks(ticks=labels_x, labels=labels_str, minor=False)
    else:
        axes.set_xticks([])

    # compute the x-label
    if (x_label is not None) and (not isinstance(x_label, str)):
        if not callable(x_label):
            raise type_error(x_label, "x_label", str, True)
        if (n_algorithms == 1) and (n_instances > 1):
            x_label = algorithm_namer(algorithms[0])
        elif (n_algorithms > 1) and (n_instances == 1):
            x_label = instance_namer(instances[0])
        else:
            x_label = x_label("algorithm_on_instance")

    z_order += 1
    pu.label_axes(axes=axes,
                  x_label=cast(str | None, x_label),
                  x_label_inside=x_label_inside,
                  x_label_location=x_label_location,
                  y_label=y_label(dimension) if callable(y_label) else y_label,
                  y_label_inside=y_label_inside,
                  y_label_location=y_label_location,
                  font_size=font_size_0,
                  z_order=z_order)

    if needs_legend:
        handles: Final[list[Line2D]] = []

        for algo in algorithms:
            linestyle = pd.create_line_style()
            linestyle["label"] = algorithm_namer(algo)
            legcol = algo_colors[algo]
            linestyle["color"] = legcol
            linestyle["markeredgecolor"] = legcol
            linestyle["xdata"] = []
            linestyle["ydata"] = []
            linestyle["linewidth"] = 6
            handles.append(Line2D(**linestyle))  # type: ignore
        z_order += 1

        axes.legend(handles=handles, loc=legend_pos,
                    labelcolor=pu.get_label_colors(handles),
                    fontsize=font_size_0).set_zorder(z_order)

    logger(f"done plotting {n_bars} end result boxes.")
    return axes



"""A multi-objective process with different search and solution spaces."""

from math import isfinite
from typing import Callable, Final

import numpy as np
from numpy import copyto
from pycommons.io.path import Path
from pycommons.types import type_error

from moptipy.api._mo_process_no_ss import _MOProcessNoSS
from moptipy.api._process_base import _TIME_IN_NS
from moptipy.api.algorithm import Algorithm
from moptipy.api.encoding import Encoding, check_encoding
from moptipy.api.logging import (
    PREFIX_SECTION_ARCHIVE,
    SCOPE_ENCODING,
    SCOPE_SEARCH_SPACE,
    SECTION_RESULT_X,
    SUFFIX_SECTION_ARCHIVE_X,
    SUFFIX_SECTION_ARCHIVE_Y,
)
from moptipy.api.mo_archive import MOArchivePruner, MORecord
from moptipy.api.mo_problem import MOProblem
from moptipy.api.space import Space, check_space
from moptipy.utils.logger import KeyValueLogSection, Logger


class _MOProcessSS(_MOProcessNoSS):
    """A class implementing a process with search and solution space."""

    def __init__(self,
                 solution_space: Space,
                 objective: MOProblem,
                 algorithm: Algorithm,
                 pruner: MOArchivePruner,
                 archive_max_size: int,
                 archive_prune_limit: int,
                 log_file: Path | None = None,
                 search_space: Space | None = None,
                 encoding: Encoding | None = None,
                 rand_seed: int | None = None,
                 max_fes: int | None = None,
                 max_time_millis: int | None = None,
                 goal_f: int | float | None = None) -> None:
        """
        Perform the internal initialization. Do not call directly.

        :param solution_space: the solution space.
        :param objective: the objective function
        :param algorithm: the optimization algorithm
        :param pruner: the archive pruner
        :param archive_max_size: the maximum archive size after pruning
        :param archive_prune_limit: the archive size above which pruning will
            be performed
        :param search_space: the search space.
        :param encoding: the encoding
        :param log_file: the optional log file
        :param rand_seed: the optional random seed
        :param max_fes: the maximum permitted function evaluations
        :param max_time_millis: the maximum runtime in milliseconds
        :param goal_f: the goal objective value. if it is reached, the
            process is terminated
        """
        super().__init__(solution_space=solution_space,
                         objective=objective,
                         algorithm=algorithm,
                         pruner=pruner,
                         archive_max_size=archive_max_size,
                         archive_prune_limit=archive_prune_limit,
                         log_file=log_file,
                         rand_seed=rand_seed,
                         max_fes=max_fes,
                         max_time_millis=max_time_millis,
                         goal_f=goal_f)

        #: The search space.
        self._search_space: Final[Space] = check_space(search_space)
        #: The encoding.
        self._encoding: Final[Encoding] = check_encoding(encoding)
        #: the internal encoder
        self._g: Final[Callable] = encoding.decode
        #: The holder for the currently de-coded solution.
        self._current_y = solution_space.create()
        #: The current best point in the search space.
        self._current_best_x: Final = search_space.create()
        # wrappers
        self.create = search_space.create  # type: ignore
        self.copy = search_space.copy  # type: ignore
        self.to_str = search_space.to_str  # type: ignore
        self.is_equal = search_space.is_equal  # type: ignore
        self.from_str = search_space.from_str  # type: ignore
        self.n_points = search_space.n_points  # type: ignore
        self.validate = search_space.validate  # type: ignore
        self._create_y = solution_space.create  # the y creator

    def f_evaluate(self, x, fs: np.ndarray) -> float | int:
        if self._terminated:
            if self._knows_that_terminated:
                raise ValueError("The process has been terminated and the "
                                 "algorithm knows it.")
            return self._current_best_f

        current_y: Final = self._current_y
        self._g(x, current_y)
        result: Final[int | float] = self._f_evaluate(current_y, fs)
        self._current_fes = current_fes = self._current_fes + 1
        do_term: bool = current_fes >= self._end_fes

        improved: bool = False
        if result < self._current_best_f:
            improved = True
            self._current_best_f = result
            copyto(self._current_best_fs, fs)
            self.copy(self._current_best_x, x)
            self._current_y = self._current_best_y
            self._current_best_y = current_y
            do_term = do_term or (result <= self._end_f)

        if self.check_in(x, fs, True) or improved:
            self._last_improvement_fe = current_fes
            self._current_time_nanos = ctn = _TIME_IN_NS()
            self._last_improvement_time_nanos = ctn

        if do_term:
            self.terminate()

        return result

    def get_copy_of_best_x(self, x) -> None:
        if self._current_fes > 0:
            return self.copy(x, self._current_best_x)
        raise ValueError("No current best x available.")

    def get_copy_of_best_y(self, y) -> None:
        if self._current_fes > 0:
            return self._copy_y(y, self._current_best_y)
        raise ValueError("No current best y available.")

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        super().log_parameters_to(logger)
        with logger.scope(SCOPE_SEARCH_SPACE) as sc:
            self._search_space.log_parameters_to(sc)
        with logger.scope(SCOPE_ENCODING) as sc:
            self._encoding.log_parameters_to(sc)

    def _write_result(self, logger: Logger) -> None:
        with logger.text(SECTION_RESULT_X) as txt:
            txt.write(self._search_space.to_str(self._current_best_x))
        super()._write_result(logger)

    def _validate_x(self) -> None:
        """Validate x, if it exists."""
        self._search_space.validate(self._current_best_x)

    def _log_and_check_archive_entry(self, index: int, rec: MORecord,
                                     logger: Logger) -> int | float:
        """
        Write an archive entry.

        :param index: the index of the entry
        :param rec: the record to verify
        :param logger: the logger
        :returns: the objective value
        """
        self.validate(rec.x)
        self.f_validate(rec.fs)
        tfs: Final[np.ndarray] = self._fs_temp

        current_y: Final = self._current_y
        self._g(rec.x, current_y)
        self._solution_space.validate(current_y)
        f: Final[int | float] = self._f_evaluate(current_y, tfs)

        if not np.array_equal(tfs, rec.fs):
            raise ValueError(
                f"expected {rec.fs} but got {tfs} when re-evaluating {rec}")
        if not isinstance(f, int | float):
            raise type_error(f, "scalarized objective value", (int, float))
        if not isfinite(f):
            raise ValueError(f"scalarized objective value {f} is not finite")

        with logger.text(f"{PREFIX_SECTION_ARCHIVE}{index}"
                         f"{SUFFIX_SECTION_ARCHIVE_X}") as lg:
            lg.write(self.to_str(rec.x))

        with logger.text(f"{PREFIX_SECTION_ARCHIVE}{index}"
                         f"{SUFFIX_SECTION_ARCHIVE_Y}") as lg:
            lg.write(self._solution_space.to_str(current_y))

        return f

    def __str__(self) -> str:
        return "MOProcessWithSearchSpace"



"""The JSSP-example specific plots."""

from math import inf
from statistics import median
from typing import Any, Callable, Final, Iterable

from pycommons.io.console import logger
from pycommons.io.path import Path, directory_path
from pycommons.types import type_error

import moptipy.utils.plot_utils as pu
from moptipy.evaluation.axis_ranger import AxisRanger
from moptipy.evaluation.base import F_NAME_RAW, F_NAME_SCALED, TIME_UNIT_MILLIS
from moptipy.evaluation.end_results import EndResult
from moptipy.evaluation.end_statistics import EndStatistics, from_end_results
from moptipy.evaluation.plot_end_results import plot_end_results
from moptipy.evaluation.plot_end_statistics_over_parameter import (
    plot_end_statistics_over_param,
)
from moptipy.evaluation.plot_progress import plot_progress
from moptipy.evaluation.progress import Progress
from moptipy.evaluation.progress import from_logs as pr_from_logs
from moptipy.evaluation.stat_run import STAT_MEAN_ARITH, StatRun
from moptipy.evaluation.stat_run import from_progress as sr_from_progress
from moptipy.examples.jssp.plot_gantt_chart import plot_gantt_chart
from moptipy.utils.lang import Lang
from moptipy.utils.plot_defaults import importance_to_font_size


def plot_end_makespans(end_results: Iterable[EndResult],
                       name_base: str,
                       dest_dir: str,
                       instance_sort_key: Callable = lambda x: x,
                       algorithm_sort_key: Callable = lambda x: x,
                       algorithm_namer: Callable[[str], str] = lambda x: x,
                       x_label_location: float = 1.0) \
        -> list[Path]:
    """
    Plot a set of end result boxes/violins functions into one chart.

    :param end_results: the iterable of end results
    :param name_base: the basic name
    :param dest_dir: the destination directory
    :param instance_sort_key: the sort key function for instances
    :param algorithm_sort_key: the sort key function for algorithms
    :param algorithm_namer: the name function for algorithms receives an
        algorithm ID and returns an instance name; default=identity function
    :param x_label_location: the location of the label of the x-axis
    :returns: the list of generated files
    """
    logger(f"beginning to plot chart {name_base}.")
    if not isinstance(end_results, Iterable):
        raise type_error(end_results, "end_results", Iterable)
    if not isinstance(name_base, str):
        raise type_error(name_base, "name_base", str)
    if not isinstance(dest_dir, str):
        raise type_error(dest_dir, "dest_dir", str)
    if not callable(instance_sort_key):
        raise type_error(instance_sort_key, "instance_sort_key", call=True)
    if not callable(algorithm_sort_key):
        raise type_error(algorithm_sort_key, "algorithm_sort_key", call=True)

    algorithms: set[str] = set()
    instances: set[str] = set()
    pairs: set[str] = set()
    for er in end_results:
        algorithms.add(er.algorithm)
        instances.add(er.instance)
        pairs.add(f"{er.algorithm}+{er.instance}")

    n_algos: Final[int] = len(algorithms)
    n_insts: Final[int] = len(instances)
    n_pairs: Final[int] = len(pairs)
    if n_pairs != (n_algos * n_insts):
        raise ValueError(
            f"found {n_algos} algorithms and {n_insts} instances, "
            f"but only {n_pairs} algorithm-instance pairs!")

    if n_algos >= 16:
        raise ValueError(f"{n_algos} are just too many algorithms...")
    max_insts: Final[int] = 16 // n_algos
    insts: Final[list[str]] = sorted(instances, key=instance_sort_key)
    result: Final[list[Path]] = []

    for lang in Lang.all_langs():
        lang.set_current()
        figure, plots = pu.create_figure_with_subplots(
            items=n_insts, max_items_per_plot=max_insts, max_rows=5,
            max_cols=1, max_width=8.6, default_height_per_row=2.5)

        for plot, start_inst, end_inst, _, _, _ in plots:
            instances.clear()
            instances.update(insts[start_inst:end_inst])
            plot_end_results(
                end_results=[er for er in end_results
                             if er.instance in instances],
                figure=plot,
                dimension=F_NAME_SCALED,
                instance_sort_key=instance_sort_key,
                algorithm_sort_key=algorithm_sort_key,
                y_label_location=1.0,
                x_label_location=x_label_location,
                algorithm_namer=algorithm_namer)

        result.extend(pu.save_figure(fig=figure,
                                     file_name=lang.filename(name_base),
                                     dir_name=dest_dir))
        del figure

    logger(f"finished plotting chart {name_base}.")
    result.sort()
    return result


def plot_stat_gantt_charts(
        end_results: Iterable[EndResult],
        results_dir: str,
        name_base: str,
        dest_dir: str,
        instance_sort_key: Callable = lambda x: x,
        statistic: Callable[[Iterable[int | float]],
                            int | float] = median) -> list[Path]:
    """
    Plot a set of Gantt charts following a specific statistics.

    :param end_results: the iterable of end results
    :param results_dir: the result directory
    :param name_base: the basic name
    :param dest_dir: the destination directory
    :param instance_sort_key: the sort key function for instances
    :param statistic: the statistic to use
    :returns: the list of generated files
    """
    logger(f"beginning to plot stat chart {name_base}.")
    if not isinstance(end_results, Iterable):
        raise type_error(end_results, "end_results", Iterable)
    if not isinstance(results_dir, str):
        raise type_error(results_dir, "results_dir", str)
    if not isinstance(name_base, str):
        raise type_error(name_base, "name_base", str)
    if not isinstance(dest_dir, str):
        raise type_error(dest_dir, "dest_dir", str)
    if not callable(instance_sort_key):
        raise type_error(instance_sort_key, "instance_sort_key", call=True)
    if not callable(statistic):
        raise type_error(statistic, "statistic", call=True)

    results: Final[list[Path]] = []  # the list of generated files

    # gather all the data
    data: Final[dict[str, list[EndResult]]] = {}
    algorithm: str | None = None
    for er in end_results:
        if algorithm is None:
            algorithm = er.algorithm
        elif algorithm != er.algorithm:
            raise ValueError(
                f"found two algorithms: {algorithm} and {er.algorithm}!")
        if er.instance in data:
            data[er.instance].append(er)
        else:
            data[er.instance] = [er]
    if algorithm is None:
        raise ValueError("Did not encounter any algorithm?")
    instances: Final[list[str]] = sorted(data.keys(), key=instance_sort_key)
    del end_results

    # get the median runs
    stat_runs: list[Path] = []
    results_dir = directory_path(results_dir)
    for instance in instances:
        runs: list[EndResult] = data[instance]
        runs.sort()
        med = statistic([er.best_f for er in runs])
        best: int | float = inf
        solution: EndResult | None = None
        for er in runs:
            current = abs(er.best_f - med)
            if current < best:
                best = current
                solution = er
        if solution is None:
            raise ValueError(
                f"found no {name_base} end result for instance {instance}.")
        path: Path = solution.path_to_file(results_dir)
        path.enforce_file()
        stat_runs.append(path)
    del data
    del instances
    if len(stat_runs) < 0:
        raise ValueError("empty set of runs?")

    # plot the gantt charts
    for lang in Lang.all_langs():
        lang.set_current()

        figure, plots = pu.create_figure_with_subplots(
            items=len(stat_runs), max_items_per_plot=1, max_cols=2,
            max_rows=4, max_width=8.6, max_height=11.5)

        for plot, start, end, _, _, _ in plots:
            if start != (end - 1):
                raise ValueError(f"{start} != {end} - 1")

            args: dict[str, Any] = {
                "gantt": stat_runs[start],
                "figure": plot,
            }
            if statistic is min:
                args["markers"] = None
            else:
                args["info"] = lambda gantt: \
                    Lang.current().format_str("gantt_info_short", gantt=gantt)
            if len(plots) > 2:
                args["importance_to_font_size_func"] = lambda i: \
                    0.9 * importance_to_font_size(i)

            plot_gantt_chart(**args)

        results.extend(pu.save_figure(fig=figure,
                                      file_name=lang.filename(name_base),
                                      dir_name=dest_dir))

    logger("done plotting stat gantt charts.")
    return results


def plot_progresses(results_dir: str,
                    algorithms: Iterable[str],
                    name_base: str,
                    dest_dir: str,
                    time_unit: str = TIME_UNIT_MILLIS,
                    log_time: bool = True,
                    instance_sort_key: Callable = lambda x: x,
                    algorithm_sort_key: Callable = lambda x: x,
                    x_label_location: float = 0.0,
                    include_runs: bool = False,
                    algorithm_namer: Callable[[str], str] = lambda x: x) \
        -> list[Path]:
    """
    Plot a set of end result boxes/violins functions into one chart.

    :param results_dir: the directory with the log files
    :param algorithms: the set of algorithms to plot together
    :param name_base: the basic name
    :param dest_dir: the destination directory
    :param time_unit: the time unit to plot
    :param log_time: should the time axis be scaled logarithmically?
    :param instance_sort_key: the sort key function for instances
    :param algorithm_sort_key: the sort key function for algorithms
    :param x_label_location: the location of the x-labels
    :param include_runs: should we include the pure runs as well?
    :param algorithm_namer: the name function for algorithms receives an
        algorithm ID and returns an instance name; default=identity function
    :returns: the list of generated files
    """
    logger(f"beginning to plot chart {name_base}.")
    if not isinstance(results_dir, str):
        raise type_error(results_dir, "results_dir", str)
    if not isinstance(algorithms, Iterable):
        raise type_error(algorithms, "algorithms", Iterable)
    if not isinstance(name_base, str):
        raise type_error(name_base, "name_base", str)
    if not isinstance(dest_dir, str):
        raise type_error(dest_dir, "dest_dir", str)
    if not isinstance(time_unit, str):
        raise type_error(time_unit, "time_unit", str)
    if not isinstance(log_time, bool):
        raise type_error(log_time, "log_time", bool)
    if not callable(instance_sort_key):
        raise type_error(instance_sort_key, "instance_sort_key", call=True)
    if not callable(algorithm_sort_key):
        raise type_error(algorithm_sort_key, "algorithm_sort_key", call=True)
    if not isinstance(x_label_location, float):
        raise type_error(x_label_location, "x_label_location", float)
    if not isinstance(include_runs, bool):
        raise type_error(include_runs, "include_runs", bool)
    if not callable(algorithm_namer):
        raise type_error(algorithm_namer, "algorithm_namer", call=True)

    # get the data
    spath: Final[Path] = directory_path(results_dir)
    progresses: Final[list[Progress]] = []
    for algorithm in sorted(algorithms, key=algorithm_sort_key):
        pr_from_logs(spath.resolve_inside(algorithm), progresses.append,
                     time_unit=time_unit, f_name=F_NAME_RAW)
    if len(progresses) <= 0:
        raise ValueError(f"did not find log files in dir {results_dir!r}.")

    stat_runs: Final[list[Progress | StatRun]] = []
    sr_from_progress(progresses, STAT_MEAN_ARITH,
                     stat_runs.append, False, False)
    if len(stat_runs) <= 0:
        raise ValueError(
            f"failed to compile stat runs from dir {results_dir!r}.")
    if include_runs:
        stat_runs.extend(progresses)
    del progresses
    instances: Final[list[str]] = sorted({sr.instance for sr in stat_runs},
                                         key=instance_sort_key)
    if len(instances) <= 0:
        raise ValueError(f"no instances in dir {results_dir!r}.")
    algos: Final[list[str]] = sorted({sr.algorithm for sr in stat_runs},
                                     key=algorithm_sort_key)
    if len(set(algorithms).difference(algos)) > 0:
        raise ValueError(
            f"found the {len(algos)} algorithms {algos}, but expected "
            f"algorithms {algorithms}.")

    results: Final[list[Path]] = []  # the list of generated files

    # plot the progress charts
    for lang in Lang.all_langs():
        lang.set_current()
        figure, plots = pu.create_figure_with_subplots(
            items=len(instances), max_items_per_plot=1, max_cols=2,
            max_rows=4, max_width=8.6, max_height=11.5)

        for plot, start, end, _, _, _ in plots:
            if start != (end - 1):
                raise ValueError(f"{start} != {end} - 1")
            inst = instances[start]
            plot_progress(
                progresses=[sr for sr in stat_runs if sr.instance == inst],
                figure=plot,
                x_axis=AxisRanger.for_axis_func(log_scale=log_time),
                importance_to_font_size_func=lambda i:
                0.9 * importance_to_font_size(i),
                algorithm_sort_key=algorithm_sort_key,
                instance_sort_key=instance_sort_key,
                x_label_location=x_label_location,
                algorithm_namer=algorithm_namer)
            axes = pu.get_axes(plot)
            pu.label_box(axes, inst, x=0.5, y=1)

        results.extend(pu.save_figure(fig=figure,
                                      file_name=lang.filename(name_base),
                                      dir_name=dest_dir))

    logger(f"finished plotting chart {name_base!r}.")
    return results


def plot_end_makespans_over_param(
    end_results: Iterable[EndResult],
    name_base: str,
    dest_dir: str,
    x_getter: Callable[[EndStatistics], int | float],
    algorithm_getter: Callable[[EndStatistics], str | None] =
    lambda es: es.algorithm,
    instance_sort_key: Callable = lambda x: x,
    algorithm_sort_key: Callable = lambda x: x,
    title: str | None = None,
    x_axis: AxisRanger | Callable[[], AxisRanger] = AxisRanger,
    x_label: str | None = None,
    x_label_location: float = 1.0,
    plot_single_instances: bool = True,
    plot_instance_summary: bool = True,
    legend_pos: str = "upper right",
    title_x: float = 0.5,
    y_label_location: float = 1.0) \
        -> list[Path]:
    """
    Plot the performance over a parameter.

    :param end_results: the iterable of end results
    :param name_base: the basic name
    :param dest_dir: the destination directory
    :param title: the optional title
    :param x_getter: the function computing the x-value for each statistics
        object
    :param algorithm_getter: the algorithm getter
    :param instance_sort_key: the sort key function for instances
    :param algorithm_sort_key: the sort key function for algorithms
    :param x_axis: the x_axis ranger
    :param x_label: the x-label
    :param x_label_location: the location of the x-labels
    :param plot_single_instances: shall we plot the values for each single
        instance?
    :param plot_instance_summary: shall we plot the value over all instances?
    :param legend_pos: the legend position
    :param title_x: the title position
    :param y_label_location: the location of the y label
    :returns: the list of generated files
    """
    logger(f"beginning to plot chart {name_base}.")
    if not isinstance(end_results, Iterable):
        raise type_error(end_results, "end_results", Iterable)
    if not isinstance(name_base, str):
        raise type_error(name_base, "name_base", str)
    if not isinstance(dest_dir, str):
        raise type_error(dest_dir, "dest_dir", str)
    if not callable(x_getter):
        raise type_error(x_getter, "x_getter", call=True)
    if not callable(algorithm_getter):
        raise type_error(algorithm_getter, "algorithm_getter", call=True)
    if not callable(instance_sort_key):
        raise type_error(instance_sort_key, "instance_sort_key", call=True)
    if not callable(algorithm_sort_key):
        raise type_error(algorithm_sort_key, "algorithm_sort_key", call=True)
    if not isinstance(x_label_location, float):
        raise type_error(x_label_location, "x_label_location", float)
    if not isinstance(y_label_location, float):
        raise type_error(y_label_location, "y_label_location", float)
    if (title is not None) and (not isinstance(title, str)):
        raise type_error(title, "title", (str, None))
    if not isinstance(plot_single_instances, bool):
        raise type_error(plot_single_instances, "plot_single_instances", bool)
    if not isinstance(plot_instance_summary, bool):
        raise type_error(plot_instance_summary, "plot_instance_summary", bool)
    if not (plot_single_instances or plot_instance_summary):
        raise ValueError("plot_instance_summary and plot_single_instances "
                         "cannot both be False")
    if not isinstance(legend_pos, str):
        raise type_error(legend_pos, "legend_pos", str)
    if not isinstance(title_x, float):
        raise type_error(title_x, "title_x", float)

    logger(f"now plotting end statistics over parameter {title!r}.")

    end_stats: Final[list[EndStatistics]] = []
    if plot_single_instances:
        from_end_results(end_results, end_stats.append)
    if plot_instance_summary:
        from_end_results(end_results, end_stats.append,
                         join_all_instances=True)
    if len(end_stats) <= 0:
        raise ValueError("no end statistics records to plot!")
    result: list[Path] = []

    for lang in Lang.all_langs():
        lang.set_current()
        figure = pu.create_figure(width=5.5)

        axes = plot_end_statistics_over_param(
            data=end_stats, figure=figure,
            algorithm_getter=algorithm_getter,
            x_axis=x_axis,
            x_getter=x_getter,
            x_label=x_label,
            x_label_location=x_label_location,
            algorithm_sort_key=algorithm_sort_key,
            instance_sort_key=instance_sort_key,
            legend_pos=legend_pos,
            y_label_location=y_label_location)
        if title is not None:
            pu.label_box(axes, title, x=title_x, y=1)

        result.extend(pu.save_figure(fig=figure,
                                     file_name=lang.filename(name_base),
                                     dir_name=dest_dir))

    logger(f"done plotting end statistics over parameter {title!r}.")
    return result



"""
The Order-based Crossover operator.

Larrañaga et al. describe this operator as follows:

The order-based crossover operator by Syswerda (1991) selects at random
several positions in a parent tour, and the order of the cities in the
selected positions of this parent is imposed on the other parent. For
example, consider the parents `12345678` and `24687531`. Suppose that in the
second parent the second, third, and sixth positions are selected. The values
in these positions are `4`, `6`, and `5` respectively. In the first parent
these cities are present at the fourth, fifth and sixth positions.
Now the offspring is equal to parent 1 except in the fourth, fifth and sixth
positions: `123xxx78`. We add the missing cities to the offspring in the same
order in which they appear in the second parent tour. This results in
`12346578`. Exchanging the role of the first parent and the second parent
gives, using the same selected positions, `24387561`.

We implement this operator such that each position has the same chance to be
chosen by either parents, i.e., the total number of positions copied from
the parents is binomially distributed with `p=0.5`, but we ensure that at
least two positions are copied from either parents (as the result would
otherwise necessarily equal one of the parents). We also switch the role of
the two parents in our implementation.

As mnemonic for the operator, we use `ox2`, similar to Larrañaga et al., who
used `OX2`.

1. G. Syswerda. Schedule Optimization Using Genetic Algorithms. In Lawrence
   Davis, L. (ed.), *Handbook of Genetic Algorithms,* pages 332-349.
   New York, NY, USA: Van Nostrand Reinhold.
2. Pedro Larrañaga, Cindy M. H. Kuijpers, Roberto H. Murga, Iñaki Inza, and
   S. Dizdarevic. Genetic Algorithms for the Travelling Salesman Problem: A
   Review of Representations and Operators. *Artificial Intelligence Review,*
   13(2):129-170, April 1999. Kluwer Academic Publishers, The Netherlands.
   https://doi.org/10.1023/A:1006529012972
"""
from typing import Callable, Final

import numpy as np
from numpy.random import Generator
from pycommons.types import type_error

from moptipy.api.operators import Op2
from moptipy.spaces.permutations import Permutations
from moptipy.utils.nputils import (
    DEFAULT_BOOL,
    DEFAULT_INT,
    fill_in_canonical_permutation,
)


# start book
class Op2OrderBased(Op2):
    """The order-based crossover operator."""

    def op2(self, random: Generator, dest: np.ndarray,
            x0: np.ndarray, x1: np.ndarray) -> None:
        """
        Apply the order-based crossover from `x0` and `x1` to `dest`.

        :param random: the random number generator
        :param dest: the array to receive the result
        :param x0: the first existing point in the search space
        :param x1: the second existing point in the search space
        """
        # end book
        indices: Final[np.ndarray] = self.__indices
        x1_done: Final[np.ndarray] = self.__x1_done
        x1_done.fill(False)  # all values in x1 are available
        rint: Final[Callable[[int], int]] = random.integers
        rbin: Final[Callable[[int, float], int]] = random.binomial
        length: Final[int] = len(indices)  # get length of string
        copy_from_x0: int  # the end index of copying from x0
        value: int  # the current value to be written to dest

        # start book
        while True:  # sample the number of values to copy from x0
            copy_from_x0 = rbin(length, 0.5)  # p=0.5 for each value
            if 1 < copy_from_x0 < (length - 1):  # ensure difference by
                break  # copying at least two values from each parent
        copy_from_x0 = length - copy_from_x0  # compute end index-index

        i: int = length  # the index into indices we iterate over
        mode: bool = True  # mode: True = copy from x0, False = from x1
        x1i: int = 0  # the index of the next unused value from x1
        while True:  # loop until we are finished
            index_i: int = rint(i)  # pick a random index-index
            index: int = indices[index_i]  # load the actual index
            i = i - 1  # reduce the number of values
            indices[i], indices[index_i] = index, indices[i]  # swap

            if mode:  # copy from x0 to dest
                dest[index] = value = x0[index]  # get and store value
                for x1j in range(x1i, length):  # mark as used
                    if (x1[x1j] == value) and (not x1_done[x1j]):
                        x1_done[x1j] = True  # mark value as used
                        break  # exit inner loop
                if copy_from_x0 == i:  # are we done with copying?
                    mode = False  # set mode to load from x1
                    x1i = 0  # reset x1 index
            else:  # go to next iteration
                dest[index] = x1[x1i]  # and store it in dest
                if i == 0:  # check if we are done
                    return  # ok, we are finished
                x1i = x1i + 1  # and move on to the next value
            while x1_done[x1i]:  # step x1i to next unused value
                x1i = x1i + 1  # increment
    # end book

    def __init__(self, space: Permutations) -> None:
        """
        Initialize the sequence crossover operator.

        :param space: the permutation space
        """
        super().__init__()
        if not isinstance(space, Permutations):
            raise type_error(space, "space", Permutations)
        if space.dimension < 4:
            raise ValueError(
                f"dimension must be > 3, but got {space.dimension}.")
        #: the valid indices
        self.__indices: Final[np.ndarray] = np.empty(
            space.dimension, DEFAULT_INT)
        #: the elements that are done in `x1`
        self.__x1_done: Final[np.ndarray] = np.ndarray(
            (space.dimension, ), DEFAULT_BOOL)

    def initialize(self) -> None:
        """Initialize this operator."""
        super().initialize()
        fill_in_canonical_permutation(self.__indices)

    def __str__(self) -> str:
        """
        Get the name of this binary operator.

        :returns: "ox2", for "order-based crossover", the name of this
            operator
        :retval "ox2": always
        """
        return "ox2"



"""Test all the links in the project's *.md files."""
# noinspection PyPackageRequirements
from pycommons.dev.tests.links_in_md import check_links_in_md
from pycommons.io.path import file_path


def test_all_links_in_readme_md() -> None:
    """Test all the links in the README.md file."""
    check_links_in_md(file_path(__file__).up(2).resolve_inside("README.md"))


def test_all_links_in_contributing_md() -> None:
    """Test all the links in the CONTRIBUTING.md file."""
    check_links_in_md(file_path(__file__).up(2).resolve_inside(
        "CONTRIBUTING.md"))


def test_all_links_in_security_md() -> None:
    """Test all the links in the SECURITY.md file."""
    check_links_in_md(file_path(__file__).up(2).resolve_inside("SECURITY.md"))


def test_all_links_in_license() -> None:
    """Test all the links in the LICENSE file."""
    check_links_in_md(file_path(__file__).up(2).resolve_inside("LICENSE"))



"""Functions for testing unary search operators with step size."""
from math import isfinite
from typing import Any, Callable, Iterable

from numpy.random import Generator, default_rng
from pycommons.types import type_error

from moptipy.api.operators import Op1WithStepSize, check_op1_with_step_size
from moptipy.api.space import Space
from moptipy.tests.op1 import default_min_unique_samples, validate_op1


def validate_op1_with_step_size(
        op1: Op1WithStepSize,
        search_space: Space | None = None,
        make_search_space_element_valid:
        Callable[[Generator, Any], Any] | None = lambda _, x: x,
        number_of_samples: int = 100,
        min_unique_samples: int | Callable[[int, Space], int]
        = default_min_unique_samples,
        step_sizes: Iterable[float] = (),
        get_step_size: Callable[[
            Space, Any, Any], float | None] | None = None) -> None:
    """
    Check whether an object is a valid moptipy unary operator with step size.

    :param op1: the operator
    :param search_space: the search space
    :param make_search_space_element_valid: make a point in the search
        space valid
    :param number_of_samples: the number of times to invoke the operator
    :param min_unique_samples: a lambda for computing the number
    :param step_sizes: the step sizes to test
    :param get_step_size: try to get the step size difference from two space
        elements
    :raises ValueError: if `op1` is not a valid instance of
        :class:`~moptipy.api.operators.Op1`
    :raises TypeError: if incorrect types are encountered
    """
    if not isinstance(op1, Op1WithStepSize):
        raise type_error(op1, "op1", Op1WithStepSize)
    if op1.__class__ == Op1WithStepSize:
        raise ValueError("Cannot use abstract base Op1WithStepSize directly.")
    check_op1_with_step_size(op1)
    if not isinstance(step_sizes, Iterable):
        raise type_error(step_sizes, "step_sizes", Iterable)
    if (get_step_size is not None) and (not callable(get_step_size)):
        raise type_error(get_step_size, "get_step_size", None, call=True)

    validate_op1(op1, search_space, make_search_space_element_valid,
                 number_of_samples, min_unique_samples)

    random: Generator | None = None
    x: Any = None
    x_copy: Any = None
    dest: Any = None
    if search_space is not None:
        if random is None:
            random = default_rng()
        x = search_space.create()
        if x is None:
            raise ValueError(
                f"search_space.create()=None for {search_space}")
        x_copy = search_space.create()
        if x_copy is None:
            raise ValueError(
                f"search_space.create()=None for {search_space}")
        dest = search_space.create()
        if dest is None:
            raise ValueError(
                f"search_space.create()=None for {search_space}")

    for i, step_size in enumerate(step_sizes):
        if not isinstance(step_size, float):
            raise type_error(step_size, f"step_sizes[{i}]", float)
        if not (isfinite(step_size) and (0 <= step_size <= 1)):
            raise ValueError(f"Forbidden step_sizes[{i}]={step_size}.")
        if (random is not None) and (search_space is not None):
            if make_search_space_element_valid is not None:
                x = make_search_space_element_valid(random, x)
            if x is None:
                raise ValueError(
                    "make_search_space_element_valid(search_space.create())="
                    f"None for {search_space}")
            search_space.validate(x)
            search_space.copy(x_copy, x)
            if not search_space.is_equal(x_copy, x):
                raise ValueError(
                    f"error when copying {x}, got {x_copy} on {search_space}")
            op1.op1(random, dest, x, step_size)
            search_space.validate(dest)
            if search_space.is_equal(dest, x_copy):
                raise ValueError(
                    f"operator copies source for step_size={step_size} "
                    f"on {search_space} and {dest}")
            if not search_space.is_equal(x, x_copy):
                raise ValueError(
                    f"operator modifies source for step_size={step_size} "
                    f"on {search_space}")
            if get_step_size is not None:
                found_step_size: float | None = \
                    get_step_size(search_space, x, dest)
                if found_step_size is None:
                    continue
                if not (isfinite(found_step_size)
                        and (0 <= found_step_size <= 1)):
                    raise ValueError(
                        f"invalid detected step size {found_step_size} "
                        f"for {search_space}.")
                if found_step_size != step_size:
                    raise ValueError(
                        f"step_size={step_size} but {op1} actually "
                        f"performs step of size {found_step_size} for "
                        f"x={x} and returns {dest} for {search_space}.")



"""An implementation of a bit string based search space."""
from typing import Final

import numpy as np
from pycommons.strings.string_conv import str_to_bool
from pycommons.types import type_error

from moptipy.spaces.nparrayspace import NPArraySpace
from moptipy.utils.nputils import DEFAULT_BOOL


class BitStrings(NPArraySpace):
    """
    A space where each element is a bit string (:class:`numpy.ndarray`).

    With such a space, discrete optimization can be realized.

    >>> s = BitStrings(5)
    >>> print(s.dimension)
    5
    >>> print(s.dtype)
    bool
    >>> print(s.create())
    [False False False False False]
    >>> print(s.to_str(s.create()))
    FFFFF
    >>> print(s.from_str(s.to_str(s.create())))
    [False False False False False]
    """

    def __init__(self, dimension: int) -> None:
        """
        Create the vector-based search space.

        :param dimension: The dimension of the search space,
            i.e., the number of decision variables.
        """
        super().__init__(dimension, DEFAULT_BOOL)

    def create(self) -> np.ndarray:
        """
        Create a bit string filled with `False`.

        :return: the string

        >>> from moptipy.spaces.bitstrings import BitStrings
        >>> s = BitStrings(8)
        >>> v = s.create()
        >>> print(s.to_str(v))
        FFFFFFFF
        >>> print(v.dtype)
        bool
        """
        return np.zeros(shape=self.dimension, dtype=DEFAULT_BOOL)

    def from_str(self, text: str) -> np.ndarray:
        """
        Convert a string to a bit string.

        :param text: the text
        :return: the vector
        :raises TypeError: if `text` is not a `str`
        :raises ValueError: if `text` cannot be converted to a valid vector
        """
        if not (isinstance(text, str)):
            raise type_error(text, "text", str)
        x: Final[np.ndarray] = self.create()
        x[:] = [str_to_bool(t) for t in text]
        self.validate(x)
        return x

    def n_points(self) -> int:
        """
        Get the scale of the bit string space.

        :return: 2 ** dimension

        >>> print(BitStrings(4).n_points())
        16
        """
        return 1 << self.dimension  # = 2 ** self.dimension

    def __str__(self) -> str:
        """
        Get the name of this space.

        :return: "bits" + dimension

        >>> print(BitStrings(5))
        bits5
        """
        return f"bits{self.dimension}"



"""
The base classes for implementing search operators.

Nullary search operators are used to sample the initial starting points of the
optimization processes. They inherit from class
:class:`~moptipy.api.operators.Op0`. The pre-defined unit test routine
:func:`~moptipy.tests.op0.validate_op0` can and should be used to test all the
nullary operators that are implemented.

Unary search operators accept one point in the search space as input and
generate a new, similar point as output. They inherit from class
:class:`~moptipy.api.operators.Op1`. The pre-defined unit test routine
:func:`~moptipy.tests.op1.validate_op1` can and should be used to test all the
unary operators that are implemented.

The basic unary operators :class:`~moptipy.api.operators.Op1` have no
parameter telling them how much of the input point to change. They may do a
hard-coded number of modifications (as, e.g.,
:class:`~moptipy.operators.permutations.op1_swap2.Op1Swap2` does) or may
apply a random number of modifications (like
:class:`~moptipy.operators.permutations.op1_swapn.Op1SwapN`). There is a
sub-class of unary operators named
:class:`~moptipy.api.operators.Op1WithStepSize` where a parameter `step_size`
with a value from the closed interval `[0.0, 1.0]` can be supplied. If
`step_size=0.0`, such an operator should perform the smallest possible
modification and for `step_size=1.0`, it should perform the largest possible
modification.

Binary search operators accept two points in the search space as input and
generate a new point that should be similar to both inputs as output. They
inherit from class :class:`~moptipy.api.operators.Op2`. The pre-defined unit
test routine :func:`~moptipy.tests.op2.validate_op2` can and should be used to
test all the binary operators that are implemented.
"""
from typing import Any

from numpy.random import Generator
from pycommons.types import type_error

from moptipy.api.component import Component


# start op0
class Op0(Component):
    """A base class to implement a nullary search operator."""

    def op0(self, random: Generator, dest) -> None:
        """
        Apply the nullary search operator to fill object `dest`.

        Afterwards `dest` will hold a valid point in the search space.
        Often, this would be a point uniformly randomly sampled from the
        search space, but it could also be the result of a heuristic or
        even a specific solution.

        :param random: the random number generator
        :param dest: the destination data structure
        """
        raise ValueError("Method not implemented!")
# end op0


def check_op0(op0: Any) -> Op0:
    """
    Check whether an object is a valid instance of :class:`Op0`.

    :param op0: the (supposed) instance of :class:`Op0`
    :return: the object `op0`
    :raises TypeError: if `op0` is not an instance of :class:`Op0`

    >>> check_op0(Op0())
    Op0
    >>> try:
    ...     check_op0('A')
    ... except TypeError as te:
    ...     print(te)
    op0 should be an instance of moptipy.api.operators.Op0 but is \
str, namely 'A'.
    >>> try:
    ...     check_op0(None)
    ... except TypeError as te:
    ...     print(te)
    op0 should be an instance of moptipy.api.operators.Op0 but is None.
    """
    if isinstance(op0, Op0):
        return op0
    raise type_error(op0, "op0", Op0)


# start op1
class Op1(Component):
    """A base class to implement a unary search operator."""

    def op1(self, random: Generator, dest, x) -> None:
        """
        Fill `dest` with a modified copy of `x`.

        :param random: the random number generator
        :param dest: the destination data structure
        :param x: the source point in the search space
        """
        raise ValueError("Method not implemented!")
# end op1


def check_op1(op1: Any) -> Op1:
    """
    Check whether an object is a valid instance of :class:`Op1`.

    :param op1: the (supposed) instance of :class:`Op1`
    :return: the object
    :raises TypeError: if `op1` is not an instance of :class:`Op1`

    >>> check_op1(Op1())
    Op1
    >>> try:
    ...     check_op1('A')
    ... except TypeError as te:
    ...     print(te)
    op1 should be an instance of moptipy.api.operators.Op1 but is str, \
namely 'A'.
    >>> try:
    ...     check_op1(None)
    ... except TypeError as te:
    ...     print(te)
    op1 should be an instance of moptipy.api.operators.Op1 but is None.
    """
    if isinstance(op1, Op1):
        return op1
    raise type_error(op1, "op1", Op1)


# start op2
class Op2(Component):
    """A base class to implement a binary search operator."""

    def op2(self, random: Generator, dest, x0, x1) -> None:
        """
        Fill `dest` with a combination of `x0` and `x1`.

        :param random: the random number generator
        :param dest: the destination data structure
        :param x0: the first source point in the search space
        :param x1: the second source point in the search space
        """
        raise ValueError("Method not implemented!")
# end op2


def check_op2(op2: Any) -> Op2:
    """
    Check whether an object is a valid instance of :class:`Op2`.

    :param op2: the (supposed) instance of :class:`Op2`
    :return: the object `op2`
    :raises TypeError: if `op2` is not an instance of :class:`Op2`

    >>> check_op2(Op2())
    Op2
    >>> try:
    ...     check_op2('A')
    ... except TypeError as te:
    ...     print(te)
    op2 should be an instance of moptipy.api.operators.Op2 but is str, \
namely 'A'.
    >>> try:
    ...     check_op2(None)
    ... except TypeError as te:
    ...     print(te)
    op2 should be an instance of moptipy.api.operators.Op2 but is None.
    """
    if isinstance(op2, Op2):
        return op2
    raise type_error(op2, "op2", Op2)


# start op1WithStepSize
class Op1WithStepSize(Op1):
    """A unary search operator with a step size."""

    def op1(self, random: Generator, dest, x, step_size: float = 0.0) -> None:
        """
        Copy `x` to `dest` but apply a modification with a given `step_size`.

        This operator is similar to :meth:`Op1.op1` in that it stores a
        modified copy of `x` into `dest`. The difference is that you can also
        specify how much that copy should be different: The parameter
        `step_size` can take on any value in the interval `[0.0, 1.0]`,
        including the two boundary values. A `step_size` of `0.0` indicates
        the smallest possible move (for which `dest` will still be different
        from `x`) and `step_size=1.0` will lead to the largest possible move.

        The `step_size` may be interpreted differently by different operators:
        Some may interpret it as an exact requirement and enforce steps of the
        exact specified size, see, for example module
        :mod:`~moptipy.operators.bitstrings.op1_flip_m`. Others might
        interpret it stochastically as an expectation. Yet others may
        interpret it as a goal step width and try to realize it in a best
        effort kind of way, but may also do smaller or larger steps if the
        best effort fails, see for example module
        :mod:`~moptipy.operators.permutations.op1_swap_exactly_n`.
        What all operators should, however, have in common is that at
        `step_size=0.0`, they should try to perform a smallest possible change
        and at `step_size=1.0`, they should try to perform a largest possible
        change. For all values in between, step sizes should grow with rising
        `step_size`. This should allow algorithms that know nothing about the
        nature of the search space or the operator's moves to still tune
        between small and large moves based on a policy which makes sense in a
        black-box setting.

        Every implementation of :class:`Op1WithStepSize` must specify a
        reasonable default value for this parameter ensure compatibility with
        :meth:`Op1.op1`. In this base class, we set the default to `0.0`.

        Finally, if a `step_size` value is passed in which is outside the
        interval `[0, 1]`, the behavior of this method is undefined. It may
        throw an exception or not. It may also enter an infinite loop.

        :param random: the random number generator
        :param dest: the destination data structure
        :param x: the source point in the search space
        :param step_size: the step size parameter for the unary operator
        """
        raise ValueError("Method not implemented!")
# end op1WithStepSize


def check_op1_with_step_size(op1: Any) -> Op1WithStepSize:
    """
    Check whether an object is a valid instance of :class:`Op1WithStepSize`.

    :param op1: the (supposed) instance of :class:`Op1WithStepSize`
    :return: the object `op1`
    :raises TypeError: if `op1` is not an instance of :class:`Op1WithStepSize`

    >>> check_op1_with_step_size(Op1WithStepSize())
    Op1WithStepSize
    >>> try:
    ...     check_op1_with_step_size('A')
    ... except TypeError as te:
    ...     print(te)
    op1 should be an instance of moptipy.api.operators.Op1WithStepSize \
but is str, namely 'A'.
    >>> try:
    ...     check_op1_with_step_size(Op1())
    ... except TypeError as te:
    ...     print(te)
    op1 should be an instance of moptipy.api.operators.Op1WithStepSize \
but is moptipy.api.operators.Op1, namely 'Op1'.
    >>> try:
    ...     check_op1_with_step_size(None)
    ... except TypeError as te:
    ...     print(te)
    op1 should be an instance of moptipy.api.operators.Op1WithStepSize \
but is None.
    """
    if isinstance(op1, Op1WithStepSize):
        return op1
    raise type_error(op1, "op1", Op1WithStepSize)



"""
An operator swapping two elements in a permutation or flipping a sign.

This operator is for :mod:`~moptipy.spaces.signed_permutations`. A similar
operator which *only* swaps elements (for :mod:`~moptipy.spaces.permutations`)
is defined in :mod:`~moptipy.operators.permutations.op1_swap2`.
"""
from typing import Callable, Final

# = import numba  # type: ignore
import numpy as np
from numpy.random import Generator

from moptipy.api.operators import Op1


# Temporary fix for https://github.com/numba/numba/issues/9103
def swap_2_or_flip(random: Generator, dest: np.ndarray,
                   x: np.ndarray) -> None:
    """
    Copy `x` into `dest` and swap two different values or flip a sign.

    :param random: the random number generator
    :param dest: the array to receive the modified copy of `x`
    :param x: the existing point in the search space

    >>> rand = np.random.default_rng(10)
    >>> xx = np.array(range(10), int)
    >>> out = np.empty(len(xx), int)
    >>> swap_2_or_flip(rand, out, xx)
    >>> print(out)
    [0 1 7 3 4 5 6 2 8 9]
    >>> swap_2_or_flip(rand, out, xx)
    >>> print(out)
    [0 1 8 3 4 5 6 7 2 9]
    >>> swap_2_or_flip(rand, out, xx)
    >>> print(out)
    [ 0  1  2  3  4 -5  6  7  8  9]
    >>> swap_2_or_flip(rand, out, xx)
    >>> print(out)
    [0 8 2 3 4 5 6 7 1 9]
    >>> swap_2_or_flip(rand, out, xx)
    >>> print(out)
    [ 0 -1  2  3  4  5  6  7  8  9]
    >>> swap_2_or_flip(rand, out, xx)
    >>> print(out)
    [ 0  1  2  3  4  5 -6  7  8  9]
    """
    dest[:] = x[:]  # First, we copy `x` to `dest`.
    length: Final[int] = len(dest)  # Get the length of `dest`.
    rint: Callable[[int, int], int] = random.integers  # fast call

    i1: Final[int] = rint(0, length)  # first random index.
    v1: Final = dest[i1]  # Get the value at the first index.

    if rint(0, 2) == 0:  # With p=0.5, we flip the sign of v1.
        dest[i1] = -v1  # Flip the sign of v1 and store it back at i1.
        return  # Quit.

# Swap two values. Now, normally we would repeat the loop until we find a
# different value. However, if we have signed permutations, it may be
# possible that all values currently are the same. To avoid an endless loop,
# we therefore use a sufficiently large range.
    for _ in range(10 + length):
        i2: int = rint(0, length)  # Get the second random index.
        v2 = dest[i2]  # Get the value at the second index.
        if v1 != v2:  # If both values different...
            dest[i2] = v1  # store v1 where v2 was
            dest[i1] = v2  # store v2 where v1 was
            return  # Exit function: we are finished.
    # If we get here, probably all elements are the same, so we just...
    dest[i1] = -v1  # ...flip the sign of v1 and store it back at i1.


class Op1Swap2OrFlip(Op1):
    """
    A search operation that swaps two (different) elements or flips a sign.

    In other words, it performs exactly one swap on a permutation or a sign
    flip. It spans a neighborhood of a rather limited size but is easy
    and fast.
    """

    def __init__(self) -> None:
        """Initialize the object."""
        super().__init__()
        self.op1 = swap_2_or_flip  # type: ignore  # use function directly

    def __str__(self) -> str:
        """
        Get the name of this unary operator.

        :returns: "swap2orFlip", the name of this operator
        :retval "swap2orFlip": always
        """
        return "swap2orFlip"



"""
A random walk algorithm implementation.

A random walk starts with a random point in the search space created with
the nullary search operator. In each step, it applies the unary search
operator to move a new point. It does not really care whether the new point
is better or worse, it will always accept it.

Of course, it still needs to call the objective function to make sure to
inform the :class:`moptipy.api.process.Process` about the new point so that,
at the end, we can obtain the best point that was visited.
But during the course of its run, it will walk around the search space
randomly without direction.
"""
from typing import Callable, Final

from numpy.random import Generator

from moptipy.api.algorithm import Algorithm1
from moptipy.api.operators import Op0, Op1
from moptipy.api.process import Process


# start book
class RandomWalk(Algorithm1):
    """
    Perform a random walk through the search space.

    In each step, a random walk creates a modified copy of the current
    solution and accepts it as starting point for the next step.
    """

    def solve(self, process: Process) -> None:
        """
        Apply the random walk to an optimization problem.

        :param process: the black-box process object
        """
        # create records for old and new point in the search space
        old_x = process.create()  # record for best-so-far solution
        new_x = process.create()  # record for new solution
        # Obtain the random number generator.
        random: Final[Generator] = process.get_random()

        # Put function references in variables to save time.
        evaluate: Final[Callable] = process.evaluate  # the objective
        op1: Final[Callable] = self.op1.op1  # the unary operator
        should_terminate: Final[Callable] = process.should_terminate

        # Start at a random point in the search space and evaluate it.
        self.op0.op0(random, new_x)  # Create one solution randomly
        evaluate(new_x)  # and evaluate it.

        while not should_terminate():  # Until we need to quit...
            old_x, new_x = new_x, old_x  # Swap old and new solution.
            op1(random, new_x, old_x)  # new_x = neighbor of old_x
            evaluate(new_x)  # Evaluate the solution and ignore result.
# end book

    def __init__(self, op0: Op0, op1: Op1) -> None:
        """
        Create the random walk.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        """
        super().__init__("rw", op0, op1)



"""
Perform restarts of an algorithm based on the Luby sequence.

Luby et al. showed that, for Las Vegas algorithms with known run time
distribution, there is an optimal stopping time in order to minimize the
expected running time. Even if the distribution is unknown, there is a
universal sequence of running times given by (1,1,2,1,1,2,4,1,1,2,1,1,2,4,
8,...), which is the optimal restarting strategy up to constant factors.
While this only holds for Las Vegas algorithms, this restart length may also
be used for optimization, e.g., if we aim to find a globally optimal
solution.

1. M. Luby, A. Sinclair, and S. Zuckerman. Optimal Speedup of Las Vegas
   Algorithms. *Information Processing Letters* 47(4):173-180. September 1993.
   https://doi.org/10.1016/0020-0190(93)90029-9

"""
from typing import Callable, Final, TypeVar, cast

import numba  # type: ignore
from pycommons.types import check_int_range, type_error

from moptipy.api.algorithm import Algorithm
from moptipy.api.mo_algorithm import MOAlgorithm
from moptipy.api.mo_process import MOProcess
from moptipy.api.process import Process
from moptipy.api.subprocesses import for_fes
from moptipy.utils.logger import CSV_SEPARATOR, KeyValueLogSection


@numba.njit(cache=True)
def _luby(i: int) -> int:
    """
    Compute the Luby sequence.

    >>> [_luby(ii) for ii in range(1, 65)]
    [1, 1, 2, 1, 1, 2, 4, 1, 1, 2, 1, 1, 2, 4, 8, 1, 1, 2, 1, 1, 2, 4, 1, \
1, 2, 1, 1, 2, 4, 8, 16, 1, 1, 2, 1, 1, 2, 4, 1, 1, 2, 1, 1, 2, 4, 8, \
1, 1, 2, 1, 1, 2, 4, 1, 1, 2, 1, 1, 2, 4, 8, 16, 32, 1]
    """
    two_by_k: int = 1
    while True:
        two_by_k_minus_one = two_by_k
        two_by_k *= 2
        if two_by_k < two_by_k_minus_one:
            return two_by_k_minus_one
        if i == (two_by_k - 1):
            return two_by_k_minus_one
        if i >= two_by_k:
            continue
        return _luby((i - two_by_k_minus_one) + 1)


class __LubyAlgorithm(Algorithm):
    """A wrapper around an existing algorithm."""

    def __init__(self, algorithm: Algorithm, base_fes: int,
                 log_restarts: bool = False) -> None:
        """
        Create the algorithm wrapper.

        :param algorithm: the algorithm to wrap
        :param base_fes: the base fes
        :param log_restarts: should we log the restarts?
        """
        super().__init__()
        #: the algorithm
        self._algo: Final[Algorithm] = algorithm
        #: the base FEs
        self._base_fes: Final[int] = base_fes
        self.initialize = algorithm.initialize  # type: ignore # fast call
        #: should we log the restarts?
        self._log_restarts: Final[bool] = log_restarts

    def __str__(self) -> str:
        """
        Get the string representation of this algorithm.

        :returns: `luby_` + sub-algorithm name
        """
        return f"luby{self._base_fes}_{self._algo}"

    def solve(self, process: Process) -> None:
        """
        Solve a single-objective problem.

        :param process: the process
        """
        st: Final[Callable[[], bool]] = process.should_terminate
        rst: Final[Callable[[], None]] = self.initialize
        sv: Final[Callable[[Process], None]] = self._algo.solve
        restarts: Final[list[tuple[int, int]] | None] = \
            [] if self._log_restarts and process.has_log() else None
        base: Final[int] = self._base_fes
        index: int = 0
        while not st():
            if (index > 1) and (restarts is not None):
                restarts.append((process.get_consumed_fes(),
                                 process.get_consumed_time_millis()))
            index = index + 1
            with for_fes(process, base * _luby(index)) as prc:
                als: Callable[[str, str], None] = prc.add_log_section

                def __als(t: str, c: str, _a=als, _i=index) -> None:
                    _a(f"{t}_{_i}", c)

                prc.add_log_section = __als  # type: ignore
                rst()
                sv(prc)
        if (restarts is not None) and (len(restarts) > 0):
            log: Final[list[str]] = [f"fes{CSV_SEPARATOR}timeMillis"]
            for row in restarts:
                log.append(CSV_SEPARATOR.join(map(
                    str, (x for x in row))))
            del restarts
            process.add_log_section("LUBY_RESTARTS", "\n".join(log))
            del log

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Store the parameters of this luby algorithm wrapper to the logger.

        :param logger: the logger
        """
        super().log_parameters_to(logger)
        logger.key_value("baseFEs", self._base_fes)
        logger.key_value("logRestarts", self._log_restarts)
        with logger.scope("a") as scope:
            self._algo.log_parameters_to(scope)


class __LubyMOAlgorithm(__LubyAlgorithm, MOAlgorithm):
    """A wrapper around an existing multi-objective algorithm."""

    def solve_mo(self, process: MOProcess) -> None:
        """
        Solve a multi-objective problem.

        :param process: the process
        """
        st: Final[Callable[[], bool]] = process.should_terminate
        rst: Final[Callable[[], None]] = self.initialize
        sv: Final[Callable[[MOProcess], None]] = cast(
            MOAlgorithm, self._algo).solve_mo
        restarts: Final[list[tuple[int, int]] | None] = \
            [] if self._log_restarts and process.has_log() else None
        base: Final[int] = self._base_fes
        index: int = 0
        while not st():
            if (index > 1) and (restarts is not None):
                restarts.append((process.get_consumed_fes(),
                                 process.get_consumed_time_millis()))
            index = index + 1
            with for_fes(process, base * _luby(index)) as prc:
                als: Callable[[str, str], None] = prc.add_log_section

                def __als(t: str, c: str, _a=als, _i=index) -> None:
                    _a(f"{t}_{_i}", c)

                prc.add_log_section = __als  # type: ignore
                rst()
                sv(prc)
        if (restarts is not None) and (len(restarts) > 0):
            log: Final[list[str]] = [f"fes{CSV_SEPARATOR}timeMillis"]
            for row in restarts:
                log.append(CSV_SEPARATOR.join(map(
                    str, (x for x in row))))
            del restarts
            process.add_log_section("LUBY_RESTARTS", "\n".join(log))
            del log

    def __str__(self) -> str:
        """
        Get the string representation of this algorithm.

        :returns: `moluby(base_fes)_` + sub-algorithm name
        """
        return f"moluby{self._base_fes}_{self._algo}"


#: the type variable for single- and multi-objective algorithms.
T = TypeVar("T", Algorithm, MOAlgorithm)


def luby_restarts(algorithm: T, base_fes: int = 64,
                  log_restarts: bool = False) -> T:
    """
    Perform restarts of an algorithm in the Luby fashion.

    The restart run length in FEs is determined by the Luby sequence (1, 1,
    2, 1, 1, 2, 4, 1, 1, 2, 1, 1, 2, 4, 8, 1, 1, 2, 1, 1, 2, 4, 1, ...)
    multiplied by `base_fes`. Restarts are performed until a termination
    criterion is met.

    :param algorithm: the algorithm
    :param base_fes: the basic objective function evaluations
    :param log_restarts: should we log the restarts?
    """
    check_int_range(base_fes, "base_fes", 1, 1_000_000_000_000)
    if not isinstance(log_restarts, bool):
        raise type_error(log_restarts, "log_restarts", bool)
    if isinstance(algorithm, MOAlgorithm):
        return __LubyMOAlgorithm(algorithm, base_fes, log_restarts)
    if isinstance(algorithm, Algorithm):
        return __LubyAlgorithm(algorithm, base_fes, log_restarts)
    raise type_error(algorithm, "algorithm", (Algorithm, MOAlgorithm))



"""Utilities for mock objects."""

from math import ceil, floor, inf, isfinite, nextafter
from typing import Callable, Final, Sequence, cast

import numpy as np
from numpy.random import Generator
from pycommons.types import type_error

#: The default types to be used for testing.
DEFAULT_TEST_DTYPES: Final[tuple[np.dtype, ...]] = tuple(sorted({
    np.dtype(bdt) for bdt in [
        int, float, np.int8, np.int16, np.uint8, np.uint16, np.int32,
        np.uint32, np.int64, np.uint64, np.float16, np.float32,
        np.float64, np.float128]}, key=lambda dt: (dt.kind, dt.itemsize)))


def _lb_int(lb: int | float) -> int:
    """
    Convert a finite lower bound to an integer.

    :param lb: the lower bound
    :retruns: the integer lower bound

    >>> _lb_int(1)
    1
    >>> type(_lb_int(1))
    <class 'int'>
    >>> _lb_int(1.4)
    2
    >>> type(_lb_int(1.4))
    <class 'int'>
    """
    return lb if isinstance(lb, int) else int(ceil(lb))


def _ub_int(ub: int | float) -> int:
    """
    Convert a finite upper bound to an integer.

    :param ub: the upper bound
    :retruns: the integer upper bound

    >>> _ub_int(1)
    1
    >>> type(_ub_int(1))
    <class 'int'>
    >>> _ub_int(1.4)
    1
    >>> type(_ub_int(1.4))
    <class 'int'>
    """
    return ub if isinstance(ub, int) else int(floor(ub))


def _float_beautify(f: float) -> float:
    """
    Get a slightly beautified float, if possible.

    :param f: the float
    :return: the beautified number
    """
    vb: int = int(round(1000.0 * f))
    r1: float = 0.001 * vb
    l1: int = len(str(r1))

    r2: float = 0.001 * (vb + 1)
    l2: int = len(str(r2))
    if l2 < l1:
        l1 = l2
        r1 = r2

    r2 = 0.001 * (vb - 1)
    l2 = len(str(r2))
    if l2 < l1:
        return r2
    return r1


def _before_int(upper_bound: int | float,
                random: Generator) -> int | None:
    """
    Get an `int` value before the given limit.

    :param upper_bound: the upper bound
    :param random: the generator
    :returns: the value, if it could be generated, `None` otherwise
    """
    if upper_bound >= inf:
        upper_bound = 10000
    elif not isfinite(upper_bound):
        return None
    lov = min(int(0.6 * upper_bound), upper_bound - 22) \
        if upper_bound > 0 else max(int(upper_bound / 0.6), upper_bound - 22)
    lo: Final[int] = _lb_int(max(lov, -9223372036854775806))
    up: Final[int] = _ub_int(upper_bound)
    if lo >= up:
        return None
    res = int(random.integers(lo, up))
    if res >= upper_bound:
        return None
    return res


def _before_float(upper_bound: int | float,
                  random: Generator) -> float | None:
    """
    Get a `float` value before the given limit.

    :param upper_bound: the upper bound
    :param random: the generator
    :returns: the value, if it could be generated, `None` otherwise
    """
    if upper_bound >= inf:
        upper_bound = 10000.0
    elif not isfinite(upper_bound):
        return None
    ulp = 1E16 * (upper_bound - nextafter(upper_bound, -inf)) \
        if (upper_bound < 0) else 1E-8 if upper_bound <= 0 \
        else 1E16 * (nextafter(upper_bound, inf) - upper_bound)
    lo = min(upper_bound * 0.6, upper_bound - ulp) \
        if upper_bound > 0.0 else min(upper_bound / 0.6, upper_bound - ulp)
    if (not isfinite(lo)) or (lo >= upper_bound):
        return None
    res = float(random.uniform(lo, upper_bound))

    if (not isfinite(res)) or (res >= upper_bound):
        return None
    resb = _float_beautify(res)
    if isfinite(resb) and (resb < upper_bound):
        return resb
    return res


def _after_int(lower_bound: int | float,
               random: Generator) -> int | None:
    """
    Get an `int` value after the given limit.

    :param lower_bound: the upper bound
    :param random: the generator
    :returns: the value, if it could be generated, `None` otherwise
    """
    if lower_bound <= -inf:
        lower_bound = -10000
    elif not isfinite(lower_bound):
        return None
    uv = max(int(lower_bound / 0.6), lower_bound + 22) \
        if lower_bound > 0 else max(int(lower_bound * 0.6), lower_bound + 22)
    ub: Final[int] = _ub_int(min(uv, 9223372036854775806))
    lb: Final[int] = _lb_int(lower_bound)
    if lb >= ub:
        return None
    res = int(random.integers(lb, ub))
    if res <= lower_bound:
        return None
    return res


def _after_float(lower_bound: int | float,
                 random: Generator) -> float | None:
    """
    Get a `float` value after the given limit.

    :param lower_bound: the upper bound
    :param random: the generator
    :returns: the value, if it could be generated, `None` otherwise
    """
    if lower_bound <= -inf:
        lower_bound = -10000.0
    elif not isfinite(lower_bound):
        return None
    ulp = 1E16 * (lower_bound - nextafter(lower_bound, -inf)) \
        if (lower_bound < 0) else 1E-8 if lower_bound <= 0 \
        else 1E16 * (nextafter(lower_bound, inf) - lower_bound)
    hi = max(lower_bound / 0.6, lower_bound + ulp) \
        if lower_bound > 0.0 else max(lower_bound * 0.6, lower_bound + ulp)
    if (not isfinite(hi)) or (hi <= lower_bound):
        return None
    res = float(random.uniform(lower_bound, hi))
    if (not isfinite(res)) or (res <= lower_bound):
        return None
    resb = _float_beautify(res)
    if isfinite(resb) and (resb > lower_bound):
        return resb
    return res


def _between_int(lower_bound: int | float,
                 upper_bound: int | float,
                 random: Generator) -> int | None:
    """
    Compute a number between two others.

    :param lower_bound: the minimum
    :param upper_bound: the maximum
    :param random: the generator
    :returns: the value, if it could be generated, `None` otherwise
    """
    if isfinite(lower_bound):
        if isfinite(upper_bound):
            lb: Final[int] = _lb_int(lower_bound) + 1
            ub: Final[int] = _ub_int(upper_bound)
            if lb < ub:
                return int(random.integers(lb, ub))
            return None
        return _after_int(lower_bound, random)
    if isfinite(upper_bound):
        return _before_int(upper_bound, random)
    return int(random.normal(0, 1000.0))


def _between_float(lower_bound: int | float,
                   upper_bound: int | float,
                   random: Generator) -> float | None:
    """
    Compute a number between two others.

    :param lower_bound: the minimum
    :param upper_bound: the maximum
    :param random: the generator
    :returns: the value, if it could be generated, `None` otherwise
    """
    if isfinite(lower_bound):
        if isfinite(upper_bound):
            a = lower_bound
            b = upper_bound
            for _ in range(5):
                a = nextafter(a, inf)
                b = nextafter(b, -inf)
            if a < b:
                res = max(a, min(b, float(random.uniform(a, b))))
                if not isfinite(res) or not (lower_bound < res < upper_bound):
                    return None
                resb = _float_beautify(res)
                if isfinite(resb) and (lower_bound < resb < upper_bound):
                    return resb
                return res
            return None
        return _after_float(lower_bound, random)
    if isfinite(upper_bound):
        return _before_float(upper_bound, random)
    return float(random.normal(0, 1000.0))


def make_ordered_list(definition: Sequence[int | float | None],
                      is_int: bool, random: Generator) \
        -> list[int | float] | None:
    """
    Make an ordered list of elements, filling in gaps.

    This function takes a list template where some values may be defined
    and some may be left `None`.
    The `None` values are then replaced such that an overall ordered list
    is created where each value is larger than its predecessor.
    The original non-`None` elements are kept in place.
    Of course, this process may fail, in which case `None` is returned.

    :param definition: a template with `None` for gaps to be filled
    :param is_int: should all generated values be integers?
    :param random: the generator
    :returns: the refined tuple with all values filled in

    >>> from numpy.random import default_rng
    >>> rg = default_rng(11)
    >>> make_ordered_list([None, 10, None, None, 50], True, rg)
    [-10, 10, 42, 47, 50]
    >>> make_ordered_list([None, 10, None, None, 50, None, None], False, rg)
    [-5.136, 10, 13.228, 15.19, 50, 115.953, 125.961]
    >>> print(make_ordered_list([9, None, 10, None, None, 50], True, rg))
    None
    >>> make_ordered_list([8, None, 10, None, None, 50], True, rg)
    [8, 9, 10, 45, 47, 50]
    >>> make_ordered_list([9, None, 10, None, None, 50], False, rg)
    [9, 9.568, 10, 47.576, 49.482, 50]
    """
    if not isinstance(definition, Sequence):
        raise type_error(definition, "definition", Sequence)
    total: Final[int] = len(definition)
    if total <= 0:
        return []

    if not isinstance(random, Generator):
        raise type_error(random, "random", Generator)
    if not isinstance(is_int, bool):
        raise type_error(is_int, "is_int", bool)

    if is_int:
        _before = cast(Callable[[int | float, Generator],
                                int | float | None], _before_int)
        _after = cast(Callable[[int | float, Generator],
                               int | float | None], _after_int)
        _between = cast(Callable[[int | float, int | float,
                                  Generator], int | float | None],
                        _between_int)
    else:
        _before = cast(Callable[[int | float, Generator],
                                int | float | None], _before_float)
        _after = cast(Callable[[int | float, Generator],
                               int | float | None], _after_float)
        _between = cast(Callable[[int | float, int | float,
                                  Generator], int | float | None],
                        _between_float)

    max_trials: int = 1000
    while max_trials > 0:
        max_trials = max_trials - 1
        result = list(definition)

        failed: bool = False

        # create one random midpoint if necessary
        has_defined: bool = False
        for i in range(total):
            if result[i] is not None:
                has_defined = True
                break
        if not has_defined:
            val = _between(-inf, inf, random)
            result[random.integers(total)] = val
            failed = val is None
        if failed:
            continue

        # fill front backwards
        for i in range(total):
            ub = result[i]
            if ub is not None:
                for j in range(i - 1, -1, -1):
                    ub = _before(ub, random)
                    if ub is None:
                        failed = True
                        break
                    result[j] = ub
                break
        if failed:
            continue

        # fill end forward
        for i in range(total - 1, -1, -1):
            lb = result[i]
            if lb is not None:
                for j in range(i + 1, total):
                    lb = _after(lb, random)
                    if lb is None:
                        failed = True
                        break
                    result[j] = lb
                break
        if failed:
            continue

        # fill all the gaps in between
        while not failed:
            # find random gap
            has_missing: bool = False
            ofs: int = random.integers(total)
            missing: int = 0
            for i in range(total):
                missing = (ofs + i) % total
                if result[missing] is None:
                    has_missing = True
                    break
            if not has_missing:
                break

            # find start of gap and lower bound
            prev_idx: int = missing
            prev: int | float | None = None
            for i in range(missing - 1, -1, -1):
                prev = result[i]
                if prev is not None:
                    prev_idx = i
                    break

            # find end of gap and upper bound
            nxt_idx: int = missing
            nxt: int | float | None = None
            for i in range(missing + 1, total):
                nxt = result[i]
                if nxt is not None:
                    nxt_idx = i
                    break

            # generate new value and store at random position in gap
            val = _between(prev, nxt, random)
            if val is None:
                failed = True
                break
            result[random.integers(prev_idx + 1, nxt_idx)] = val

        if failed:
            continue

        # now check and return result
        prev = result[0]
        if prev is None:
            continue
        for i in range(1, total):
            nxt = result[i]
            if (nxt is None) or (nxt <= prev):
                failed = True
                break
            prev = nxt
        if not failed:
            return result

    return None


def sample_from_attractors(random: Generator,
                           attractors: Sequence[int | float],
                           is_int: bool = False,
                           lb: int | float = -inf,
                           ub: int | float = inf) -> int | float:
    """
    Sample from a given range using the specified attractors.

    :param random: the random number generator
    :param attractors: the attractor points
    :param lb: the lower bound
    :param ub: the upper bound
    :param is_int: shall we sample integer values?
    :return: the value
    :raises ValueError: if the sampling failed

    >>> from numpy.random import default_rng
    >>> rg = default_rng(11)
    >>> sample_from_attractors(rg, [5, 20])
    15.198106552324713
    >>> sample_from_attractors(rg, [2], lb=0, ub=10, is_int=True)
    3
    >>> sample_from_attractors(rg, [5, 20], lb=4)
    4.7448464616061665
    >>> sample_from_attractors(rg, [5, 20], ub=22)
    1.044618552249311
    >>> sample_from_attractors(rg, [5, 20], lb=0, ub=30, is_int=True)
    6
    >>> sample_from_attractors(rg, [5, 20], lb=4, ub=22, is_int=True)
    20
    """
    max_trials: int = 1000
    al: Final[int] = len(attractors)
    while max_trials > 0:
        max_trials -= 1

        chosen_idx = random.integers(al)
        chosen = attractors[chosen_idx]
        lo = attractors[chosen_idx - 1] if (chosen_idx > 0) else lb
        hi = attractors[chosen_idx + 1] if (chosen_idx < (al - 1)) else ub

        sd = 0.5 * min(hi - chosen, chosen - lo)
        if not isfinite(sd):
            sd = max(1.0, 0.05 * abs(chosen))
        sample = random.normal(chosen, sd)
        if not isfinite(sample):
            continue
        sample = int(sample) if is_int else float(sample)
        if lb <= sample <= ub:
            return sample

    raise ValueError(f"Failed to sample with lb={lb}, ub={ub}, "
                     f"attractors={attractors}, is_int={is_int}.")
